# Brian Heim
# 2018-08-26
#
# See accompanying README for info.

import argparse
import os
import sys
import shutil
import getpass
import re
import datetime
from functools import reduce

# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument('root', nargs = '?', help = "Root project directory, where CMakeLists.txt will be written. Defaults to current working directory", default = '.')
parser.add_argument('-p', '--plugins', nargs = '+', help = "Plugin subproject directories, for example -p Sin Saw Square")
parser.add_argument('-P', '--project-name', help = "Name of the project. Defaults to the name of the root project directory")
parser.add_argument('-c', '--min-cmake-version', help = "Mininum CMake version to require.  Defaults to 3.12, must be >= 3.12", default = '3.12')
parser.add_argument('-s', '--cpp-standard', help = "C++ standard to use. Defaults to C++17.  Valid arguments are 11, 14, 17.", type = int, default = 17)
parser.add_argument('-v', '--verbose', help = "Verbose output. Can be specified up to 2 times for more verbosity", action = 'count', default = 0)
parser.add_argument('-a', '--author', help = "Project author name.")
parser.add_argument('-D', '--date', help = "Project date. Defaults to now.")
parser.add_argument('-f', '--flat-dir', help = "Use flat-directory configuration. See README for more info.")
parser.add_argument('-C', '--sc-files-dir', help = "For use with flat-directory configuration.  Directory where SC files can be found. Defaults to --flat-dir argument.")
parser.add_argument('-H', '--schelp-files-dir', help = "For use with flat-directory configuration.  Directory where schelp files can be found. Defaults to --flat-dir argument.")
parser.add_argument('-I', '--install-cmake', action = 'store_true',
        help = "Install CMake modules after running. root/cmake_modules is created if necessary.")
parser.add_argument('--install-cmake-only', action = 'store_true',
        help = "Install CMake modules without running generation script.")

MODULES = [
    'SuperColliderCompilerConfig.cmake',
    'SuperColliderServerPlugin.cmake'
];
HEADER_TEMPLATE = 'CMakeLists_header.template'
FOOTER_TEMPLATE = 'CMakeLists_footer.template'
TARGET_TEMPLATE_TEXT = '''
####################################################################################################
# Begin target {name}

set({name}_cpp_files
{cpp})
set({name}_sc_files
{sc})
set({name}_schelp_files
{schelp})

sc_add_server_plugin(
    "{project}/{name}" # desination directory
    "{name}" # target name
    "${{{name}_cpp_files}}"
    "${{{name}_sc_files}}"
    "${{{name}_schelp_files}}"
)

# End target {name}
####################################################################################################
'''

# script output
DEBUG = 2
VERBOSE = 1
NORMAL = 0
WARNING = -1
ERROR = -2

verbosity_ = 0

def out(s, level):
    if verbosity_ >= level:
        if level == WARNING:
            print('WARNING: ' + s)
        elif level == ERROR:
            print('ERROR: ' + s)
        else:
            print(s)

# convenience FP utils
def apply(f, l):
    for x in l:
        f(x)

def mapnow(f, l):
    return [f(x) for x in l]

def filternow(f, l):
    return list(filter(f, l))

def zip2now(a, b):
    return list(zip(a, b))

# main generation utility class
class Generator:
    def __init__(self, args):
        global verbosity_
        verbosity_ = args.verbose
        self.args = args
        self.script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
        self.root_dir = os.path.realpath(args.root)
        self.project_name_spaceless = self.args.project_name.replace(' ', '')
        if not args_.install_cmake_only:
            self.validate()

    def validate(self):
        flatmode = self.args.flat_dir or self.args.sc_files_dir or self.args.schelp_files_dir
        if flatmode and self.args.plugins:
            out('Cannot specify both flat-dir mode and individual plugin directories', ERROR)
            exit(1)

        if not flatmode:
            if not self.args.plugins:
                out('Must specify one of -p or -f', ERROR)
                parser.print_help()
                exit(1)
            else:
                for plugin in self.args.plugins:
                    plugin_path = os.path.join(self.root_dir, plugin)
                    if not os.path.exists(plugin_path):
                        out('Config: Plugin directory does not exist: {}'.format(plugin_path), ERROR)
                        exit(1)

        else:
            self.validate_flatmode()

        out('Config: Directory of this script: ' + self.script_dir, DEBUG)
        out('Config: Absolute path of root directory: ' + self.root_dir, DEBUG)

        if not self.args.cpp_standard in [11, 14, 17]:
            out('Invalid C++ standard given.', ERROR)
            exit(1)
        if not os.path.exists(self.root_dir):
            out('Root project directory does not exist: {}'.format(self.root_dir), ERROR)
            exit(1)

        if not int(self.args.min_cmake_version[0:1]) == 3 or not float(self.args.min_cmake_version[2:]) >= 12.0:
            out('CMake version must be >= 3.12, got {}'.format(self.args.min_cmake_version))
            exit(1)
        if not self.args.author:
            user = Generator.get_username()
            out('Config: No author given, using user profile name: {}'.format(user), NORMAL)
            self.args.author = user
        if not self.args.date:
            date = '{:%Y-%m-%d}'.format(datetime.datetime.now())
            out('Config: No date given, using now: {}'.format(date), VERBOSE)
            self.args.date = date

        if not self.args.project_name:
            self.args.project_name = os.path.split(self.root_dir)[1]
            out('Config: No project name given, using parent of root dirctory: {}'.format(self.args.project_name), VERBOSE)

    def validate_flatmode(self):
        if not self.args.sc_files_dir:
            self.args.sc_files_dir = self.args.flat_dir
        if not self.args.schelp_files_dir:
            self.args.schelp_files_dir = self.args.flat_dir

        self.full_flat_dir = os.path.join(self.root_dir, self.args.flat_dir)
        self.full_sc_files_dir = os.path.join(self.root_dir, self.args.sc_files_dir)
        self.full_schelp_files_dir = os.path.join(self.root_dir, self.args.schelp_files_dir)
        out('Config: sc_files_dir is {}'.format(self.full_sc_files_dir), VERBOSE)
        out('Config: schelp_files_dir is {}'.format(self.full_schelp_files_dir), VERBOSE)
        def check_exists(name, dir):
            if not os.path.exists(dir):
                out('{} does not exist: {}'.format(name, dir), ERROR)
                exit(1)

        check_exists('--flat-dir', self.full_flat_dir)
        check_exists('--sc-files-dir', self.full_sc_files_dir)
        check_exists('--schelp-files-dir', self.full_schelp_files_dir)


    # main generation routine
    def generate_cmake(self):
        if self.args.flat_dir:
            targets, project_files = self.find_project_files_flat(self.full_flat_dir)
        else:
            targets, project_files = self.find_project_files(self.args.plugins)
        out('Targets: {}; project_files: {}'.format(targets, project_files), DEBUG)
        result = self.gen_section('header', HEADER_TEMPLATE)
        result += self.gen_targets(targets, project_files)
        result += self.gen_section('footer', FOOTER_TEMPLATE)
        self.write_cmake_file(result)

    def gen_section(self, name, filename):
        out('Generating {}'.format(name), VERBOSE)
        path = os.path.join(self.script_dir, filename)
        out('Reading from {} file {}'.format(name, path), DEBUG)
        with open(path, mode='r') as f:
            section = ''.join(f.readlines())
        out('Configuring {}'.format(name), DEBUG)
        return self.configure_text(section)

    def write_cmake_file(self, text):
        cmake_file = os.path.join(self.root_dir, 'CMakeLists.txt')
        out('Writing CMake file'.format(cmake_file), DEBUG)
        with open(cmake_file, mode='w') as f:
            f.write(text)
        out('Wrote project file to {}'.format(cmake_file), NORMAL)

    def gen_targets(self, plugins, project_files):
        out('Generating targets text', VERBOSE)
        return reduce(lambda x, y: x + y,
            map(lambda p: self.gen_target_cmake(p,
                project_files[p]['cpp'],
                project_files[p]['sc'],
                project_files[p]['schelp']
                ),
                plugins
            )
        )

    # replace '@xyz' variables in template text with script arguments
    def configure_text(self, text):
        envir = [
            ('project_name', self.args.project_name),
            ('author', self.args.author),
            ('minimum_cmake_version', self.args.min_cmake_version),
            ('cpp_standard', str(self.args.cpp_standard)),
            ('date', self.args.date)
        ]

        out('Configure environment: {}'.format(envir), DEBUG)

        def op(text, pair):
            k, v = pair
            k = '@' + k
            out('\tReplacement: {} -> {}'.format(k, v, text), DEBUG)
            return re.sub(k, v, text)

        return reduce(op, envir, text)

    def gen_target_cmake(self, name, cpp_files, sc_files, schelp_files):
        out('\tGenerating CMake for plugin {}'.format(name), VERBOSE)
        result = TARGET_TEMPLATE_TEXT.format(
            project=self.project_name_spaceless,
            name=name,
            cpp=self.join_file_list(cpp_files),
            sc=self.join_file_list(sc_files),
            schelp=self.join_file_list(schelp_files)
        )

        return result

    # helper
    def join_file_list(self, files):
        if files:
            # CMake generally likes / more than \
            return ''.join(map(lambda x: '    ' + x.replace('\\', '/') + '\n', files))
        else:
            return ''

    # return type is a tuple of plugin names and a map from plugin name to a map from 'sc', 'cpp',
    # 'schelp' to the relevant files for that target. 'cpp' will map to both header and
    # implementation files.
    def find_project_files(self, dirs):
        out('Scanning for project files.', VERBOSE)
        out('Directories: {}'.format(dirs), DEBUG)

        # split will return (path, '') if path ends in dir separator
        def get_real_parent(path):
            head, tail = os.path.split(path)
            if tail == '':
                return os.path.split(head)[1]
            else:
                return tail

        plugin_names = mapnow(get_real_parent, dirs)
        result = dict()
        for d, p in zip(dirs, plugin_names):
            result[p] = self.find_project_files_impl(d)

        return (plugin_names, result)

    # Helper for find_project_files
    def find_project_files_impl(self, dir):
        dirmap = dict(cpp=[], sc=[], schelp=[])
        d = os.path.join(self.root_dir, dir)
        out('Current directory: {}'.format(d), DEBUG)

        def classify(f):
            ty = Generator.classify_file_type(f)
            out('\tClassified {} as {}'.format(f, ty), DEBUG)
            if ty:
                fullpath = os.path.join(root, f)
                relpath = os.path.relpath(fullpath, self.root_dir)
                out('\tIncluding {} as a {} file'.format(relpath, ty), VERBOSE)
                dirmap[ty].append(relpath)

        for root, _, files in os.walk(d):
            apply(classify, files)

        return dirmap

    # Gather all .cpp files in a directory, absolute paths
    def get_cpp_files(self, dir):
        entries = os.listdir(dir)
        cppfiles = zip2now(entries, map(lambda f: os.path.join(dir, f), entries))
        cppfiles = filternow(lambda f: os.path.isfile(f[1]), cppfiles)
        cppfiles = filternow(lambda f: Generator.is_cpp_file(f[1]), cppfiles)
        cppfiles = sorted(cppfiles)
        return cppfiles

    # Same as find_project_files, but with flat directories
    def find_project_files_flat(self, fulldir):
        out('Scanning for project files in dir: {}'.format(fulldir), VERBOSE)
        result = dict()
        plugins = []
        cppfiles = self.get_cpp_files(fulldir)

        def scan_for_filetype(dir, pluginname, ext, test):
            out('Scanning for matching .{} files in dir: {}'.format(ext, dir), DEBUG)
            for root, _, files in os.walk(dir):
                for file in files:
                    if test(file) and os.path.splitext(file)[0] == pluginname:
                        relpath = os.path.relpath(os.path.join(root, file), self.root_dir)
                        out('\tIncluding {} in target'.format(relpath), VERBOSE)
                        result[pluginname][ext].append(relpath)

        for fname, fpath in cppfiles:
            pluginname, _ = os.path.splitext(fname)
            out('Found plugin {}'.format(pluginname), NORMAL)
            plugins.append(pluginname)

            result[pluginname] = dict(cpp=[], sc=[], schelp=[])
            relpath = os.path.relpath(fpath, self.root_dir)
            out('\tIncluding {} as a cpp file'.format(relpath), DEBUG)
            result[pluginname]['cpp'].append(relpath)

            scan_for_filetype(self.full_flat_dir, pluginname, 'cpp', Generator.is_hpp_file)

        # Find all [sc, schelp] files in dir. Match and remove files that share a name with known
        # targets.  Generate warnings for the remaining files; as this signals some plugins may not
        # have been found.
        def find_matching_files(dir, ext, test):
            extfiles = []
            for root, _, files in os.walk(dir):
                extfiles.extend(mapnow(lambda f: os.path.join(root, f), filternow(test, files)))

            for plugin in plugins:
                extname = plugin + '.' + ext
                for extfile in extfiles:
                    if os.path.split(extfile)[1] == extname:
                        relpath = os.path.relpath(extfile, self.root_dir)
                        out('\tIncluding {} in target'.format(relpath), VERBOSE)
                        result[plugin][ext].append(relpath)
                        extfiles.remove(extfile)
                        break

            for leftover in extfiles:
                out('File {} was not included in any target'.format(leftover), WARNING)

        find_matching_files(self.full_sc_files_dir, 'sc', Generator.is_sc_file)
        find_matching_files(self.full_schelp_files_dir, 'schelp', Generator.is_schelp_file)

        return (plugins, result)

    # filetype tests
    @staticmethod
    def is_cpp_file(name):
        return any(map(lambda x: name.endswith(x), ['.cpp', '.cxx', '.c']))

    @staticmethod
    def is_hpp_file(name):
        return any(map(lambda x: name.endswith(x), ['.hpp', '.hxx', '.h']))

    @staticmethod
    def is_sc_file(name):
        return name.endswith('.sc')

    @staticmethod
    def is_schelp_file(name):
        return name.endswith('.schelp')

    @staticmethod
    def classify_file_type(name):
        if Generator.is_sc_file(name):
            return 'sc'
        if Generator.is_schelp_file(name):
            return 'schelp'
        if Generator.is_cpp_file(name) or Generator.is_hpp_file(name):
            return 'cpp'

        return None

    @staticmethod
    def get_username():
        return getpass.getuser()

    def install_modules(self):
        out('Installing CMake modules', DEBUG)
        modules_dir = os.path.join(g.root_dir, 'cmake_modules')

        try:
            os.mkdir(modules_dir)
            out('Created cmake_modules directory {}.'.format(modules_dir), DEBUG)
        except FileExistsError:
            out('cmake_modules directory {} already existed'.format(modules_dir), DEBUG)

        for module in MODULES:
            pfrom = os.path.join(self.script_dir, module)
            if not os.path.exists(pfrom):
                out('Module does not exist: {}'.format(pfrom), ERROR)
                continue

            out('Installing {} to {}'.format(pfrom, modules_dir), VERBOSE)
            shutil.copy(pfrom, modules_dir)

        out('Installed {} CMake modules'.format(len(MODULES)), NORMAL)

if __name__ == "__main__":
    args_ = parser.parse_args()
    g = Generator(args_)
    if not args_.install_cmake_only:
        g.generate_cmake()

    if args_.install_cmake or args_.install_cmake_only:
        g.install_modules()



"""Server info UGens."""

from ..synthdef import UGen, param, ugen


@ugen(ir=True)
class BlockSize(UGen):
    pass


@ugen(kr=True, ir=True)
class BufChannels(UGen):
    buffer_id = param()


@ugen(kr=True, ir=True)
class BufDur(UGen):
    buffer_id = param()


@ugen(kr=True, ir=True)
class BufFrames(UGen):
    buffer_id = param()


@ugen(kr=True, ir=True)
class BufRateScale(UGen):
    buffer_id = param()


@ugen(kr=True, ir=True)
class BufSampleRate(UGen):
    buffer_id = param()


@ugen(kr=True, ir=True)
class BufSamples(UGen):
    buffer_id = param()


@ugen(ir=True)
class ControlDur(UGen):
    pass


@ugen(ir=True)
class ControlRate(UGen):
    pass


@ugen(ir=True)
class NodeID(UGen):
    pass


@ugen(ir=True)
class NumAudioBuses(UGen):
    pass


@ugen(ir=True)
class NumBuffers(UGen):
    pass


@ugen(ir=True)
class NumControlBuses(UGen):
    pass


@ugen(ir=True)
class NumInputBuses(UGen):
    pass


@ugen(ir=True)
class NumOutputBuses(UGen):
    pass


@ugen(kr=True, ir=True)
class NumRunningSynths(UGen):
    pass


@ugen(ir=True)
class RadiansPerSample(UGen):
    pass


@ugen(ir=True)
class SampleDur(UGen):
    pass


@ugen(ir=True)
class SampleRate(UGen):
    pass


@ugen(ir=True)
class SubsampleOffset(UGen):
    pass



"""
07_delay_reverb.py -- Delay lines and reverb.

Demonstrates effect processing with three SynthDefs:
  1. A short percussive source (filtered saw impulse)
  2. Ping-pong style comb delay feedback
  3. FreeVerb for lush tail

The source feeds into the comb delay, which feeds into reverb,
showing how to chain effects via audio buses.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import AddAction, Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import (
    AllpassC,
    CombC,
    FreeVerb,
    In,
    Out,
    Pan2,
    RLPF,
    Saw,
)


# First private audio bus (after hardware I/O buses)
EFFECT_BUS = 16


def main() -> None:
    # -- SynthDef 1: percussive source -> effect bus --------------------------
    with SynthDefBuilder(frequency=440.0, amplitude=0.4) as builder:
        sig = Saw.ar(frequency=builder["frequency"])
        sig = RLPF.ar(source=sig, frequency=2000.0, reciprocal_of_q=0.3)
        env = EnvGen.kr(
            envelope=Envelope.percussive(attack_time=0.003, release_time=0.15),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        # Write to both hardware out (dry) and effect bus
        Out.ar(bus=0, source=Pan2.ar(source=sig * 0.4))
        Out.ar(bus=EFFECT_BUS, source=sig)

    src_def = builder.build(name="perc_src")
    print(f"SynthDef '{src_def.name}' compiled: {len(src_def.compile())} bytes")

    # -- SynthDef 2: comb delay effect (reads effect bus) ---------------------
    with SynthDefBuilder(delay_time=0.375, decay_time=3.0, mix=0.5) as builder:
        dry = In.ar(bus=EFFECT_BUS)
        # Two comb delays at slightly different times for stereo width
        left = CombC.ar(
            source=dry,
            maximum_delay_time=1.0,
            delay_time=builder["delay_time"],
            decay_time=builder["decay_time"],
        )
        right = CombC.ar(
            source=dry,
            maximum_delay_time=1.0,
            delay_time=builder["delay_time"] * 0.75,
            decay_time=builder["decay_time"],
        )
        # Allpass diffusion
        left = AllpassC.ar(
            source=left,
            maximum_delay_time=0.05,
            delay_time=0.031,
            decay_time=1.0,
        )
        right = AllpassC.ar(
            source=right,
            maximum_delay_time=0.05,
            delay_time=0.043,
            decay_time=1.0,
        )
        wet = builder["mix"]
        Out.ar(bus=0, source=[left * wet, right * wet])

    delay_def = builder.build(name="comb_delay")
    print(f"SynthDef '{delay_def.name}' compiled: {len(delay_def.compile())} bytes")

    # -- SynthDef 3: reverb tail (reads hardware out, adds reverb) ------------
    with SynthDefBuilder(room=0.85, damp=0.4, mix=0.3) as builder:
        sig = In.ar(bus=0, channel_count=2)
        left = sig[0]
        right = sig[1]
        left = FreeVerb.ar(
            source=left,
            mix=builder["mix"],
            room_size=builder["room"],
            damping=builder["damp"],
        )
        right = FreeVerb.ar(
            source=right,
            mix=builder["mix"],
            room_size=builder["room"],
            damping=builder["damp"],
        )
        Out.ar(bus=0, source=[left, right])

    verb_def = builder.build(name="reverb")
    print(f"SynthDef '{verb_def.name}' compiled: {len(verb_def.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        src_def.send(server)
        delay_def.send(server)
        verb_def.send(server)
        time.sleep(0.1)

        # Create source and effect groups with correct execution order
        src_group = server.group(target=1, action=AddAction.ADD_TO_HEAD)
        fx_group = server.group(target=int(src_group), action=AddAction.ADD_AFTER)

        # Start persistent effect synths in the effect group
        server.synth(
            "comb_delay",
            target=int(fx_group),
            delay_time=0.375,
            decay_time=4.0,
            mix=0.4,
        )
        server.synth(
            "reverb",
            target=int(fx_group),
            action=AddAction.ADD_AFTER,  # after fx group, not inside it
            room=0.8,
            damp=0.5,
            mix=0.25,
        )

        # Fire percussive notes in the source group
        melody = [
            (329.63, 0.375),  # E4
            (392.00, 0.375),  # G4
            (440.00, 0.375),  # A4
            (523.25, 0.375),  # C5
            (440.00, 0.375),  # A4
            (392.00, 0.375),  # G4
            (329.63, 0.750),  # E4 (held longer)
            (293.66, 0.750),  # D4
        ]

        print("Playing percussive melody through delay + reverb...")
        for freq, dur in melody:
            server.synth(
                "perc_src",
                target=int(src_group),
                frequency=freq,
                amplitude=0.5,
            )
            time.sleep(dur)

        # Let the delay tail ring out
        print("Letting delay tail ring out (4s)...")
        time.sleep(4.0)

    print("Done.")


if __name__ == "__main__":
    main()



"""Chaotic generator UGens."""

from ..synthdef import UGen, param, ugen


@ugen(ar=True)
class CuspL(UGen):
    frequency = param(22050)
    a = param(1.0)
    b = param(1.9)
    xi = param(0.0)


@ugen(ar=True)
class CuspN(UGen):
    frequency = param(22050)
    a = param(1.0)
    b = param(1.9)
    xi = param(0.0)


@ugen(ar=True)
class FBSineC(UGen):
    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class FBSineL(UGen):
    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class FBSineN(UGen):
    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class GbmanL(UGen):
    frequency = param(22050)
    xi = param(1.2)
    yi = param(2.1)


@ugen(ar=True)
class GbmanN(UGen):
    frequency = param(22050)
    xi = param(1.2)
    yi = param(2.1)


@ugen(ar=True)
class HenonC(UGen):
    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class HenonL(UGen):
    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class HenonN(UGen):
    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class LatoocarfianC(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LatoocarfianL(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LatoocarfianN(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LinCongC(UGen):
    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LinCongL(UGen):
    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LinCongN(UGen):
    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LorenzL(UGen):
    frequency = param(22050)
    s = param(10)
    r = param(28)
    b = param(2.667)
    h = param(0.05)
    xi = param(0.1)
    yi = param(0)
    zi = param(0)


@ugen(ar=True)
class QuadC(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class QuadL(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class QuadN(UGen):
    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class StandardL(UGen):
    frequency = param(22050)
    k = param(1)
    xi = param(0.5)
    yi = param(0)


@ugen(ar=True)
class StandardN(UGen):
    frequency = param(22050)
    k = param(1)
    xi = param(0.5)
    yi = param(0)



"""
08_resonance.py -- Resonant textures with Dust and Ringz.

Sparse random impulses (Dust) excite banks of resonant filters (Ringz)
at different frequencies, creating a shimmering bell/chime texture.
Demonstrates trigger-driven synthesis and resonant filter banks.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import Dust, HPF, LeakDC, Out, Pan2, Ringz


def main() -> None:
    # -- SynthDef: resonant chime bank ----------------------------------------
    with SynthDefBuilder(density=1.5, decay=2.0, amplitude=0.3) as builder:
        trig = Dust.ar(density=builder["density"])
        decay = builder["decay"]

        # Bank of resonant filters at harmonically-related frequencies
        # (tuned to an E minor chord spread across octaves)
        sig = Ringz.ar(source=trig, frequency=329.63, decay_time=decay) * 0.15
        sig = (
            sig + Ringz.ar(source=trig, frequency=493.88, decay_time=decay * 0.8) * 0.12
        )
        sig = (
            sig + Ringz.ar(source=trig, frequency=659.26, decay_time=decay * 0.7) * 0.10
        )
        sig = (
            sig + Ringz.ar(source=trig, frequency=987.77, decay_time=decay * 0.5) * 0.08
        )
        sig = (
            sig + Ringz.ar(source=trig, frequency=1318.5, decay_time=decay * 0.4) * 0.06
        )
        sig = (
            sig + Ringz.ar(source=trig, frequency=1975.5, decay_time=decay * 0.3) * 0.04
        )

        # Remove DC offset and sub-bass rumble
        sig = LeakDC.ar(source=sig)
        sig = HPF.ar(source=sig, frequency=200.0)

        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=1.0, sustain_time=5.0, release_time=2.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    synthdef = builder.build(name="chime_bank")
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        synthdef.send(server)
        time.sleep(0.1)

        # Layer two instances panned apart for stereo spread
        print("Playing resonant chime texture (8s)...")
        server.synth("chime_bank", density=1.2, decay=2.5, amplitude=0.4)
        time.sleep(0.5)
        server.synth("chime_bank", density=0.8, decay=3.0, amplitude=0.3)
        time.sleep(8.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""
11_stereo_pad.py -- Lush detuned stereo pad with MoogFF filter.

Builds a thick stereo pad by layering detuned Pulse oscillators with
varying widths, run through a MoogFF (Moog-style ladder filter) with
an LFO-modulated cutoff. Demonstrates:
  - Detuning for stereo width
  - Pulse width as a timbral control
  - MoogFF resonant filter
  - Slow LFO modulation at control rate
  - Multi-voice layering within a single SynthDef

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFNoise2, MoogFF, Out, Pulse


def main() -> None:
    # -- SynthDef: detuned stereo pad with MoogFF -----------------------------
    with SynthDefBuilder(
        frequency=220.0,
        detune=0.8,
        cutoff=1200.0,
        resonance=2.5,
        amplitude=0.3,
        gate=1.0,
    ) as builder:
        freq = builder["frequency"]
        det = builder["detune"]

        # Left channel: two detuned pulse oscillators
        left = Pulse.ar(frequency=freq - det, width=0.3) * 0.5
        left = left + Pulse.ar(frequency=freq - det * 2.0, width=0.5) * 0.5

        # Right channel: two detuned pulse oscillators (opposite detune)
        right = Pulse.ar(frequency=freq + det, width=0.3) * 0.5
        right = right + Pulse.ar(frequency=freq + det * 2.0, width=0.5) * 0.5

        # Slowly wandering cutoff modulation
        lfo = LFNoise2.kr(frequency=0.2)
        cutoff = builder["cutoff"] + lfo * 600.0

        # MoogFF ladder filter on each channel
        left = MoogFF.ar(source=left, frequency=cutoff, gain=builder["resonance"])
        right = MoogFF.ar(source=right, frequency=cutoff, gain=builder["resonance"])

        # ADSR envelope with gate control
        env = EnvGen.kr(
            envelope=Envelope.adsr(
                attack_time=1.5,
                decay_time=0.5,
                sustain=0.8,
                release_time=2.0,
            ),
            gate=builder["gate"],
            done_action=DoneAction.FREE_SYNTH,
        )
        left = left * env * builder["amplitude"]
        right = right * env * builder["amplitude"]
        Out.ar(bus=0, source=[left, right])

    synthdef = builder.build(name="stereo_pad")
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        synthdef.send(server)
        time.sleep(0.1)

        # Play a chord progression: Am -> F -> C -> G
        chords = [
            (220.00, "Am"),  # A3
            (174.61, "F"),  # F3
            (261.63, "C"),  # C4
            (196.00, "G"),  # G3
        ]

        print("Playing pad chord progression...")
        for freq, name in chords:
            print(f"  {name} ({freq:.0f} Hz)...")
            node = server.synth(
                "stereo_pad",
                frequency=freq,
                detune=0.7,
                cutoff=1000.0,
                resonance=2.0,
                amplitude=0.25,
            )
            time.sleep(3.0)
            # Release the gate
            node.set(gate=0.0)
            time.sleep(1.0)

        # Let the last release tail finish
        time.sleep(2.0)

    print("Done.")


if __name__ == "__main__":
    main()



"""
13_server_pad.py -- Gated pad with managed_synth.

Demonstrates managed_synth() with a gated ADSR envelope. Each chord in
the progression is a managed_synth context -- the node is freed on exit,
guaranteeing cleanup even if an exception occurs mid-performance.

Also demonstrates:
  - Server context manager for boot/quit lifecycle
  - SynthDef.send() / SynthDef.play()
  - server.set() for parameter updates on a running node
  - server.managed_group() for grouping voices

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFNoise2, LPF, Out, Saw, SinOsc


def main() -> None:
    # -- SynthDef 1: gated saw pad with filter LFO ----------------------------
    with SynthDefBuilder(
        frequency=220.0,
        cutoff=1200.0,
        amplitude=0.25,
        gate=1.0,
    ) as builder:
        sig = Saw.ar(frequency=builder["frequency"])
        # Slow filter drift
        lfo = LFNoise2.kr(frequency=0.3)
        cutoff = builder["cutoff"] + lfo * 400.0
        sig = LPF.ar(source=sig, frequency=cutoff)
        env = EnvGen.kr(
            envelope=Envelope.adsr(
                attack_time=0.8,
                decay_time=0.3,
                sustain=0.6,
                release_time=1.5,
            ),
            gate=builder["gate"],
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=[sig, sig])

    pad_def = builder.build(name="pad")
    print(f"SynthDef '{pad_def.name}' compiled: {len(pad_def.compile())} bytes")
    print(pad_def.dump_ugens())

    # -- SynthDef 2: simple sine bell (self-freeing) --------------------------
    with SynthDefBuilder(frequency=880.0, amplitude=0.15) as builder:
        sig = SinOsc.ar(frequency=builder["frequency"])
        env = EnvGen.kr(
            envelope=Envelope.percussive(attack_time=0.005, release_time=0.8),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=[sig, sig])

    bell_def = builder.build(name="bell")

    # -- Chord progression ----------------------------------------------------
    chords = [
        ("Am", [220.00, 261.63, 329.63]),
        ("F", [174.61, 220.00, 261.63]),
        ("C", [261.63, 329.63, 392.00]),
        ("G", [196.00, 246.94, 293.66]),
    ]

    with Server(Options(verbosity=0, load_synthdefs=False)) as server:
        pad_def.send(server)
        bell_def.send(server)
        time.sleep(0.1)

        print("\nPlaying chord progression with managed_synth...")

        for chord_name, freqs in chords:
            print(f"  {chord_name}...")

            # Create a managed group for this chord's voices
            with server.managed_group(target=1) as group:
                # Spawn one pad voice per note in the chord
                nodes = []
                for freq in freqs:
                    node = server.synth(
                        "pad",
                        target=group,
                        action=0,
                        frequency=freq,
                        cutoff=1000.0,
                        amplitude=0.15,
                    )
                    nodes.append(node)

                # Accent bell on the root
                bell_def.play(server, frequency=freqs[0] * 2, amplitude=0.1)

                # Let the chord sustain
                time.sleep(2.5)

                # Slowly brighten the filter before releasing
                for node in nodes:
                    server.set(node, cutoff=2000.0)
                time.sleep(0.5)

                # Release gates for clean ADSR release tails
                for node in nodes:
                    server.set(node, gate=0.0)
                time.sleep(1.5)

            # Group freed here -- any leftover voices are cleaned up

        # Brief silence before exit
        time.sleep(0.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""Tests for NodeProxy and Ndef (live coding proxy system).

All tests use a mock server to avoid booting audio.
Source callables use real UGens (they compile without a running server).
"""

from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import pytest

from nanosynth.proxy import Ndef, NodeProxy
from nanosynth.scsynth import BootStatus
from nanosynth.server import Server
from nanosynth.synthdef import SynthDefBuilder
from nanosynth.ugens.osc import SinOsc
from nanosynth.ugens.inout import Out


def _sine_source() -> Any:
    """A minimal UGen callable for testing."""
    return SinOsc.ar(frequency=440)  # type: ignore[attr-defined]


@pytest.fixture()
def server() -> Server:
    """Server with mocked protocol (no audio engine needed)."""
    s = Server()
    s._protocol = MagicMock()
    s._protocol.status = BootStatus.ONLINE
    return s


# ---------------------------------------------------------------------------
# NodeProxy
# ---------------------------------------------------------------------------


class TestNodeProxy:
    def test_initial_state(self, server: Server) -> None:
        proxy = NodeProxy(server)
        assert proxy.bus is None
        assert proxy.source is None
        assert proxy.is_playing is False

    def test_bus_allocated_on_source_set(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        assert proxy.bus is not None
        assert proxy.bus.num_channels == 2

    def test_source_setter_with_callable(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        assert proxy.source is _sine_source
        # Verify synthdef was sent and synth was created
        assert server._protocol.send_packet.call_count >= 2  # d_recv + s_new

    def test_source_setter_with_synthdef(self, server: Server) -> None:
        with SynthDefBuilder(out=0.0) as builder:
            Out.ar(bus=builder["out"], source=SinOsc.ar())  # type: ignore[attr-defined]
        sd = builder.build(name="test_source")

        proxy = NodeProxy(server)
        proxy.source = sd
        assert proxy.source is sd

    def test_source_setter_none_clears(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        proxy.source = None
        assert proxy.source is None

    def test_source_swap_frees_old(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        first_call_count = server._protocol.send_packet.call_count

        # Swap source -- should send gate=0 for old + d_recv + s_new for new
        proxy.source = _sine_source
        assert server._protocol.send_packet.call_count > first_call_count

    def test_version_increments(self, server: Server) -> None:
        proxy = NodeProxy(server)
        assert proxy._version == 0
        proxy.source = _sine_source
        assert proxy._version == 1
        proxy.source = _sine_source
        assert proxy._version == 2

    def test_play_creates_monitor(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        initial_count = server._protocol.send_packet.call_count

        proxy.play()
        assert proxy.is_playing is True
        # Should have sent d_recv (monitor synthdef) + s_new (monitor synth)
        assert server._protocol.send_packet.call_count > initial_count

    def test_play_idempotent(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        proxy.play()
        count_after_first = server._protocol.send_packet.call_count

        proxy.play()  # second call should be no-op
        assert server._protocol.send_packet.call_count == count_after_first

    def test_stop_clears_monitor(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        proxy.play()
        assert proxy.is_playing is True

        proxy.stop()
        assert proxy.is_playing is False

    def test_stop_when_not_playing_is_noop(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.stop()  # should not raise

    def test_clear_frees_everything(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        proxy.play()

        proxy.clear()
        assert proxy.source is None
        assert proxy.bus is None
        assert proxy.is_playing is False

    def test_set_params(self, server: Server) -> None:
        proxy = NodeProxy(server)
        proxy.source = _sine_source
        proxy.set(frequency=880.0)
        # The set should have been sent (last send_packet call)
        assert server._protocol.send_packet.call_count > 0

    def test_set_without_source_raises(self, server: Server) -> None:
        proxy = NodeProxy(server)
        with pytest.raises(RuntimeError, match="No source synth"):
            proxy.set(frequency=440.0)

    def test_invalid_source_type_raises(self, server: Server) -> None:
        proxy = NodeProxy(server)
        with pytest.raises(TypeError, match="callable, SynthDef, or None"):
            proxy.source = 42  # type: ignore[assignment]

    def test_custom_channel_count(self, server: Server) -> None:
        proxy = NodeProxy(server, num_channels=4)
        proxy.source = _sine_source
        assert proxy.bus is not None
        assert proxy.bus.num_channels == 4


# ---------------------------------------------------------------------------
# Ndef
# ---------------------------------------------------------------------------


class TestNdef:
    def setup_method(self) -> None:
        """Clear the registry before each test."""
        Ndef._registry.clear()

    def test_creates_proxy(self, server: Server) -> None:
        proxy = Ndef(server, "test")
        assert isinstance(proxy, NodeProxy)

    def test_returns_same_proxy(self, server: Server) -> None:
        p1 = Ndef(server, "test")
        p2 = Ndef(server, "test")
        assert p1 is p2

    def test_different_names_different_proxies(self, server: Server) -> None:
        p1 = Ndef(server, "a")
        p2 = Ndef(server, "b")
        assert p1 is not p2

    def test_sets_source(self, server: Server) -> None:
        proxy = Ndef(server, "test", _sine_source)
        assert proxy.source is _sine_source

    def test_hot_swap(self, server: Server) -> None:
        def other_source() -> Any:
            return SinOsc.ar(frequency=880)  # type: ignore[attr-defined]

        Ndef(server, "test", _sine_source)
        proxy = Ndef(server, "test", other_source)
        assert proxy.source is other_source

    def test_no_source_arg_does_not_overwrite(self, server: Server) -> None:
        Ndef(server, "test", _sine_source)
        proxy = Ndef(server, "test")
        assert proxy.source is _sine_source

    def test_clear_all(self, server: Server) -> None:
        Ndef(server, "a", _sine_source)
        Ndef(server, "b", _sine_source)
        Ndef.clear_all(server)
        assert len(Ndef._registry) == 0

    def test_clear_all_scoped_to_server(self) -> None:
        s1 = Server()
        s1._protocol = MagicMock()
        s1._protocol.status = BootStatus.ONLINE

        s2 = Server()
        s2._protocol = MagicMock()
        s2._protocol.status = BootStatus.ONLINE

        Ndef(s1, "a", _sine_source)
        Ndef(s2, "a", _sine_source)

        Ndef.clear_all(s1)
        # s2's proxy should still be in the registry
        assert any(k[0] == id(s2) for k in Ndef._registry)
        assert not any(k[0] == id(s1) for k in Ndef._registry)

    def test_source_none_clears(self, server: Server) -> None:
        Ndef(server, "test", _sine_source)
        proxy = Ndef(server, "test", None)
        assert proxy.source is None



"""UGen definitions for nanosynth, organized by category."""

from .basic import (
    Mix,
    MulAdd,
    Sum3,
    Sum4,
)
from .beq import (
    BAllPass,
    BBandPass,
    BBandStop,
    BHiCut,
    BHiPass,
    BHiShelf,
    BLowCut,
    BLowPass,
    BLowShelf,
    BPeakEQ,
)
from .bufio import (
    BufRd,
    BufWr,
    ClearBuf,
    LocalBuf,
    MaxLocalBufs,
    PlayBuf,
    RecordBuf,
    ScopeOut,
    ScopeOut2,
)
from .chaos import (
    CuspL,
    CuspN,
    FBSineC,
    FBSineL,
    FBSineN,
    GbmanL,
    GbmanN,
    HenonC,
    HenonL,
    HenonN,
    LatoocarfianC,
    LatoocarfianL,
    LatoocarfianN,
    LinCongC,
    LinCongL,
    LinCongN,
    LorenzL,
    QuadC,
    QuadL,
    QuadN,
    StandardL,
    StandardN,
)
from .convolution import (
    Convolution,
    Convolution2,
    Convolution2L,
    Convolution3,
)
from .delay import (
    AllpassC,
    AllpassL,
    AllpassN,
    BufAllpassC,
    BufAllpassL,
    BufAllpassN,
    BufCombC,
    BufCombL,
    BufCombN,
    BufDelayC,
    BufDelayL,
    BufDelayN,
    CombC,
    CombL,
    CombN,
    DelTapRd,
    DelTapWr,
    Delay1,
    Delay2,
    DelayC,
    DelayL,
    DelayN,
)
from .demand import (
    Dbrown,
    Dbufrd,
    Dbufwr,
    Demand,
    DemandEnvGen,
    Dgeom,
    Dibrown,
    Diwhite,
    Drand,
    Dreset,
    Dseq,
    Dser,
    Dseries,
    Dshuf,
    Dstutter,
    Dswitch,
    Dswitch1,
    Dunique,
    Duty,
    Dwhite,
    Dwrand,
    Dxrand,
)
from .diskio import (
    DiskIn,
    DiskOut,
    VDiskIn,
)
from .dynamics import (
    Amplitude,
    Compander,
    CompanderD,
    Limiter,
    Normalizer,
)
from .envelopes import (
    Done,
    Free,
    FreeSelf,
    FreeSelfWhenDone,
    Linen,
    Pause,
    PauseSelf,
    PauseSelfWhenDone,
)
from .ffsinosc import (
    Blip,
    FSinOsc,
    Klank,
    Pulse,
    Saw,
)
from .filters import (
    APF,
    BPF,
    BPZ2,
    BRF,
    BRZ2,
    Changed,
    Decay,
    Decay2,
    DetectSilence,
    FOS,
    Formlet,
    HPF,
    HPZ1,
    HPZ2,
    Integrator,
    LPF,
    LPZ1,
    LPZ2,
    Lag,
    Lag2,
    Lag2UD,
    Lag3,
    Lag3UD,
    LagUD,
    LeakDC,
    Median,
    MidEQ,
    MoogFF,
    OnePole,
    OneZero,
    RHPF,
    RLPF,
    Ramp,
    Ringz,
    SOS,
    Slew,
    Slope,
    TwoPole,
    TwoZero,
)
from .gendyn import (
    Gendy1,
    Gendy2,
    Gendy3,
)
from .granular import (
    GrainBuf,
    GrainIn,
    PitchShift,
    Warp1,
)
from .hilbert import (
    FreqShift,
    Hilbert,
    HilbertFIR,
)
from .info import (
    BlockSize,
    BufChannels,
    BufDur,
    BufFrames,
    BufRateScale,
    BufSampleRate,
    BufSamples,
    ControlDur,
    ControlRate,
    NodeID,
    NumAudioBuses,
    NumBuffers,
    NumControlBuses,
    NumInputBuses,
    NumOutputBuses,
    NumRunningSynths,
    RadiansPerSample,
    SampleDur,
    SampleRate,
    SubsampleOffset,
)
from .inout import (
    In,
    InFeedback,
    LocalIn,
    LocalOut,
    OffsetOut,
    Out,
    ReplaceOut,
    XOut,
)
from .lines import (
    A2K,
    AmpComp,
    AmpCompA,
    DC,
    K2A,
    LinExp,
    LinLin,
    Line,
    Silence,
    XLine,
)
from .mac import (
    KeyState,
    MouseButton,
    MouseX,
    MouseY,
)
from .ml import (
    BeatTrack,
    BeatTrack2,
    KeyTrack,
    Loudness,
    MFCC,
    Onsets,
    Pitch,
    SpecCentroid,
    SpecFlatness,
    SpecPcile,
)
from .noise import (
    BrownNoise,
    ClipNoise,
    CoinGate,
    Crackle,
    Dust,
    Dust2,
    ExpRand,
    GrayNoise,
    Hasher,
    IRand,
    LFClipNoise,
    LFDClipNoise,
    LFDNoise0,
    LFDNoise1,
    LFDNoise3,
    LFNoise0,
    LFNoise1,
    LFNoise2,
    LinRand,
    Logistic,
    MantissaMask,
    NRand,
    PinkNoise,
    Rand,
    RandID,
    RandSeed,
    TExpRand,
    TIRand,
    TRand,
    TWindex,
    WhiteNoise,
)
from .osc import (
    COsc,
    DegreeToKey,
    Impulse,
    Index,
    LFCub,
    LFGauss,
    LFPar,
    LFPulse,
    LFSaw,
    LFTri,
    Osc,
    OscN,
    Select,
    SinOsc,
    SyncSaw,
    VOsc,
    VOsc3,
    VarSaw,
    Vibrato,
    WrapIndex,
)
from .panning import (
    Balance2,
    BiPanB2,
    DecodeB2,
    Pan2,
    Pan4,
    PanAz,
    PanB,
    PanB2,
    Rotate2,
    Splay,
    XFade2,
)
from .physical import (
    Ball,
    Pluck,
    Spring,
    TBall,
)
from .pv import (
    FFT,
    IFFT,
    PV_Add,
    PV_BinScramble,
    PV_BinShift,
    PV_BinWipe,
    PV_BrickWall,
    PV_ChainUGen,
    PV_ConformalMap,
    PV_Conj,
    PV_Copy,
    PV_CopyPhase,
    PV_Diffuser,
    PV_Div,
    PV_HainsworthFoote,
    PV_JensenAndersen,
    PV_LocalMax,
    PV_MagAbove,
    PV_MagBelow,
    PV_MagClip,
    PV_MagDiv,
    PV_MagFreeze,
    PV_MagMul,
    PV_MagNoise,
    PV_MagShift,
    PV_MagSmear,
    PV_MagSquared,
    PV_Max,
    PV_Min,
    PV_Mul,
    PV_PhaseShift,
    PV_PhaseShift270,
    PV_PhaseShift90,
    PV_RandComb,
    PV_RandWipe,
    PV_RectComb,
    PV_RectComb2,
    RunningSum,
)
from .reverb import FreeVerb
from .safety import (
    CheckBadValues,
    Sanitize,
)
from .triggers import (
    Clip,
    Fold,
    Gate,
    InRange,
    Latch,
    LeastChange,
    MostChange,
    Peak,
    PeakFollower,
    Phasor,
    Poll,
    RunningMax,
    RunningMin,
    Schmidt,
    SendPeakRMS,
    SendReply,
    SendTrig,
    Sweep,
    TDelay,
    ToggleFF,
    Trig,
    Trig1,
    Wrap,
    ZeroCrossing,
)

__all__ = [
    # basic
    "Mix",
    "MulAdd",
    "Sum3",
    "Sum4",
    # beq
    "BAllPass",
    "BBandPass",
    "BBandStop",
    "BHiCut",
    "BHiPass",
    "BHiShelf",
    "BLowCut",
    "BLowPass",
    "BLowShelf",
    "BPeakEQ",
    # bufio
    "BufRd",
    "BufWr",
    "ClearBuf",
    "LocalBuf",
    "MaxLocalBufs",
    "PlayBuf",
    "RecordBuf",
    "ScopeOut",
    "ScopeOut2",
    # chaos
    "CuspL",
    "CuspN",
    "FBSineC",
    "FBSineL",
    "FBSineN",
    "GbmanL",
    "GbmanN",
    "HenonC",
    "HenonL",
    "HenonN",
    "LatoocarfianC",
    "LatoocarfianL",
    "LatoocarfianN",
    "LinCongC",
    "LinCongL",
    "LinCongN",
    "LorenzL",
    "QuadC",
    "QuadL",
    "QuadN",
    "StandardL",
    "StandardN",
    # convolution
    "Convolution",
    "Convolution2",
    "Convolution2L",
    "Convolution3",
    # delay
    "AllpassC",
    "AllpassL",
    "AllpassN",
    "BufAllpassC",
    "BufAllpassL",
    "BufAllpassN",
    "BufCombC",
    "BufCombL",
    "BufCombN",
    "BufDelayC",
    "BufDelayL",
    "BufDelayN",
    "CombC",
    "CombL",
    "CombN",
    "DelTapRd",
    "DelTapWr",
    "Delay1",
    "Delay2",
    "DelayC",
    "DelayL",
    "DelayN",
    # demand
    "Dbrown",
    "Dbufrd",
    "Dbufwr",
    "Demand",
    "DemandEnvGen",
    "Dgeom",
    "Dibrown",
    "Diwhite",
    "Drand",
    "Dreset",
    "Dseq",
    "Dser",
    "Dseries",
    "Dshuf",
    "Dstutter",
    "Dswitch",
    "Dswitch1",
    "Dunique",
    "Duty",
    "Dwhite",
    "Dwrand",
    "Dxrand",
    # diskio
    "DiskIn",
    "DiskOut",
    "VDiskIn",
    # dynamics
    "Amplitude",
    "Compander",
    "CompanderD",
    "Limiter",
    "Normalizer",
    # envelopes
    "Done",
    "Free",
    "FreeSelf",
    "FreeSelfWhenDone",
    "Linen",
    "Pause",
    "PauseSelf",
    "PauseSelfWhenDone",
    # ffsinosc
    "Blip",
    "FSinOsc",
    "Klank",
    "Pulse",
    "Saw",
    # filters
    "APF",
    "BPF",
    "BPZ2",
    "BRF",
    "BRZ2",
    "Changed",
    "Decay",
    "Decay2",
    "DetectSilence",
    "FOS",
    "Formlet",
    "HPF",
    "HPZ1",
    "HPZ2",
    "Integrator",
    "LPF",
    "LPZ1",
    "LPZ2",
    "Lag",
    "Lag2",
    "Lag2UD",
    "Lag3",
    "Lag3UD",
    "LagUD",
    "LeakDC",
    "Median",
    "MidEQ",
    "MoogFF",
    "OnePole",
    "OneZero",
    "RHPF",
    "RLPF",
    "Ramp",
    "Ringz",
    "SOS",
    "Slew",
    "Slope",
    "TwoPole",
    "TwoZero",
    # gendyn
    "Gendy1",
    "Gendy2",
    "Gendy3",
    # granular
    "GrainBuf",
    "GrainIn",
    "PitchShift",
    "Warp1",
    # hilbert
    "FreqShift",
    "Hilbert",
    "HilbertFIR",
    # info
    "BlockSize",
    "BufChannels",
    "BufDur",
    "BufFrames",
    "BufRateScale",
    "BufSampleRate",
    "BufSamples",
    "ControlDur",
    "ControlRate",
    "NodeID",
    "NumAudioBuses",
    "NumBuffers",
    "NumControlBuses",
    "NumInputBuses",
    "NumOutputBuses",
    "NumRunningSynths",
    "RadiansPerSample",
    "SampleDur",
    "SampleRate",
    "SubsampleOffset",
    # inout
    "In",
    "InFeedback",
    "LocalIn",
    "LocalOut",
    "OffsetOut",
    "Out",
    "ReplaceOut",
    "XOut",
    # lines
    "A2K",
    "AmpComp",
    "AmpCompA",
    "DC",
    "K2A",
    "LinExp",
    "LinLin",
    "Line",
    "Silence",
    "XLine",
    # mac
    "KeyState",
    "MouseButton",
    "MouseX",
    "MouseY",
    # ml
    "BeatTrack",
    "BeatTrack2",
    "KeyTrack",
    "Loudness",
    "MFCC",
    "Onsets",
    "Pitch",
    "SpecCentroid",
    "SpecFlatness",
    "SpecPcile",
    # noise
    "BrownNoise",
    "ClipNoise",
    "CoinGate",
    "Crackle",
    "Dust",
    "Dust2",
    "ExpRand",
    "GrayNoise",
    "Hasher",
    "IRand",
    "LFClipNoise",
    "LFDClipNoise",
    "LFDNoise0",
    "LFDNoise1",
    "LFDNoise3",
    "LFNoise0",
    "LFNoise1",
    "LFNoise2",
    "LinRand",
    "Logistic",
    "MantissaMask",
    "NRand",
    "PinkNoise",
    "Rand",
    "RandID",
    "RandSeed",
    "TExpRand",
    "TIRand",
    "TRand",
    "TWindex",
    "WhiteNoise",
    # osc
    "COsc",
    "DegreeToKey",
    "Impulse",
    "Index",
    "LFCub",
    "LFGauss",
    "LFPar",
    "LFPulse",
    "LFSaw",
    "LFTri",
    "Osc",
    "OscN",
    "Select",
    "SinOsc",
    "SyncSaw",
    "VOsc",
    "VOsc3",
    "VarSaw",
    "Vibrato",
    "WrapIndex",
    # panning
    "Balance2",
    "BiPanB2",
    "DecodeB2",
    "Pan2",
    "Pan4",
    "PanAz",
    "PanB",
    "PanB2",
    "Rotate2",
    "Splay",
    "XFade2",
    # physical
    "Ball",
    "Pluck",
    "Spring",
    "TBall",
    # pv
    "FFT",
    "IFFT",
    "PV_Add",
    "PV_BinScramble",
    "PV_BinShift",
    "PV_BinWipe",
    "PV_BrickWall",
    "PV_ChainUGen",
    "PV_ConformalMap",
    "PV_Conj",
    "PV_Copy",
    "PV_CopyPhase",
    "PV_Diffuser",
    "PV_Div",
    "PV_HainsworthFoote",
    "PV_JensenAndersen",
    "PV_LocalMax",
    "PV_MagAbove",
    "PV_MagBelow",
    "PV_MagClip",
    "PV_MagDiv",
    "PV_MagFreeze",
    "PV_MagMul",
    "PV_MagNoise",
    "PV_MagShift",
    "PV_MagSmear",
    "PV_MagSquared",
    "PV_Max",
    "PV_Min",
    "PV_Mul",
    "PV_PhaseShift",
    "PV_PhaseShift270",
    "PV_PhaseShift90",
    "PV_RandComb",
    "PV_RandWipe",
    "PV_RectComb",
    "PV_RectComb2",
    "RunningSum",
    # reverb
    "FreeVerb",
    # safety
    "CheckBadValues",
    "Sanitize",
    # triggers
    "Clip",
    "Fold",
    "Gate",
    "InRange",
    "Latch",
    "LeastChange",
    "MostChange",
    "Peak",
    "PeakFollower",
    "Phasor",
    "Poll",
    "RunningMax",
    "RunningMin",
    "Schmidt",
    "SendPeakRMS",
    "SendReply",
    "SendTrig",
    "Sweep",
    "TDelay",
    "ToggleFF",
    "Trig",
    "Trig1",
    "Wrap",
    "ZeroCrossing",
]



#!/usr/bin/python

# Copyright (C) 2003-2017 Erik de Castro Lopo <erikd@mega-nerd.com>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the author nor the names of any contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re, sys

#----------------------------------------------------------------
# These are all of the public functions exported from libsndfile.
#
# Its important not to change the order they are listed in or
# the ordinal values in the second column.

ALL_SYMBOLS = (
	(	"sf_command",	 		1	),
	(	"sf_open",				2	),
	(	"sf_close",				3	),
	(	"sf_seek",				4	),
	(	"sf_error",				7	),
	(	"sf_perror",			8	),
	(	"sf_error_str",			9	),
	(	"sf_error_number",		10	),
	(	"sf_format_check",		11	),
	(	"sf_read_raw",			16	),
	(	"sf_readf_short",		17	),
	(	"sf_readf_int",			18	),
	(	"sf_readf_float",		19	),
	(	"sf_readf_double",		20	),
	(	"sf_read_short",		21	),
	(	"sf_read_int",			22	),
	(	"sf_read_float",		23	),
	(	"sf_read_double",		24	),
	(	"sf_write_raw",			32	),
	(	"sf_writef_short",		33	),
	(	"sf_writef_int",		34	),
	(	"sf_writef_float",		35	),
	(	"sf_writef_double",		36	),
	(	"sf_write_short",		37	),
	(	"sf_write_int",			38	),
	(	"sf_write_float",		39	),
	(	"sf_write_double",		40	),
	(	"sf_strerror",			50	),
	(	"sf_get_string",		60	),
	(	"sf_set_string",		61	),
	(	"sf_version_string",	68	),
	(	"sf_open_fd",			70	),
	(	"sf_wchar_open",		71  ),
	(	"sf_open_virtual",		80	),
	(	"sf_write_sync",		90	),
	(	"sf_set_chunk",			100	),
	(	"sf_get_chunk_size",	101 ),
	(	"sf_get_chunk_data",	102 ),
	(	"sf_get_chunk_iterator",	103 ),
	(	"sf_next_chunk_iterator",	104 ),
	(	"sf_current_byterate",	110 )
	)

#-------------------------------------------------------------------------------

def linux_symbols (progname, version):
	print ("# Auto-generated by %s\n" %progname)
	print ("libsndfile.so.%s" % version)
	print ("{")
	print ("  global:")
	for name, ordinal in ALL_SYMBOLS:
		if  name == "sf_wchar_open":
			continue
		print ("    %s ;" % name)
	print ("  local:")
	print ("    * ;")
	print ("} ;")
	sys.stdout.write ("\n")
	return

def darwin_symbols (progname, version):
	print ("# Auto-generated by %s\n" %progname)
	for name, ordinal in ALL_SYMBOLS:
		if  name == "sf_wchar_open":
			continue
		print ("_%s" % name)
	sys.stdout.write ("\n")
	return

def win32_symbols (progname, version, name):
	print ("; Auto-generated by %s\n" %progname)
	print ("EXPORTS\n")
	for name, ordinal in ALL_SYMBOLS:
		print ("%-20s @%s" % (name, ordinal))
	sys.stdout.write ("\n")
	return

def os2_symbols (progname, version, name):
	print ("; Auto-generated by %s\n" %progname)
	print ("LIBRARY %s%s" % (name, re.sub (r"\..*", "", version)))
	print ("INITINSTANCE TERMINSTANCE")
	print ("CODE PRELOAD MOVEABLE DISCARDABLE")
	print ("DATA PRELOAD MOVEABLE MULTIPLE NONSHARED")
	print ("EXPORTS\n")
	for name, ordinal in ALL_SYMBOLS:
		if  name == "sf_wchar_open":
			continue
		print ("_%-20s @%s" % (name, ordinal))
	sys.stdout.write ("\n")
	return

def plain_symbols (progname, version, name):
	for name, ordinal in ALL_SYMBOLS:
		print (name)

def no_symbols (os_name):
	sys.stdout.write ("\n")
	print ("No known way of restricting exported symbols on '%s'." % os_name)
	print ("If you know a way, please contact the author.")
	sys.stdout.write ("\n")
	return

#-------------------------------------------------------------------------------

progname = re.sub (".*[\\/]", "", sys.argv [0])

if len (sys.argv) != 3:
	sys.stdout.write ("\n")
	print ("Usage : %s <target OS name> <libsndfile version>." % progname)
	sys.stdout.write ("\n")
	print ("    Currently supported values for target OS are:")
	print ("          linux")
	print ("          darwin     (ie MacOSX)")
	print ("          win32      (ie wintendo)")
	print ("          cygwin     (Cygwin on wintendo)")
	print ("          os2        (OS/2)")
	print ("          plain      (plain list of symbols)")
	sys.stdout.write ("\n")
	sys.exit (1)

os_name = sys.argv [1]
version = re.sub (r"\.[a-z0-9]+$", "", sys.argv [2])

if os_name == "linux" or os_name == "gnu" or os_name == "binutils":
	linux_symbols (progname, version)
elif os_name == "darwin":
	darwin_symbols (progname, version)
elif os_name == "win32":
	win32_symbols (progname, version, "libsndfile")
elif os_name == "cygwin":
	win32_symbols (progname, version, "cygsndfile")
elif os_name == "os2":
	os2_symbols (progname, version, "sndfile")
elif os_name == "static":
	plain_symbols (progname, version, "")
else:
	no_symbols (os_name)

sys.exit (0)




"""
04_additive.py -- Additive synthesis.

Builds a tone by summing sine partials with decreasing amplitudes.
An LFO slowly modulates the balance between odd and even harmonics,
creating a timbral shift from hollow (odd-only, like a square wave)
to bright (all harmonics).

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFTri, Out, Pan2, SinOsc


def main() -> None:
    # -- SynthDef: additive with LFO-morphing harmonics -----------------------
    with SynthDefBuilder(frequency=200.0, amplitude=0.3) as builder:
        freq = builder["frequency"]

        # LFO: 0..1 controls even-harmonic level (odd harmonics always present)
        even_mix = LFTri.kr(frequency=0.15) * 0.5 + 0.5

        # Odd harmonics (1, 3, 5, 7)
        sig = SinOsc.ar(frequency=freq) * 1.0
        sig = sig + SinOsc.ar(frequency=freq * 3.0) * 0.33
        sig = sig + SinOsc.ar(frequency=freq * 5.0) * 0.2
        sig = sig + SinOsc.ar(frequency=freq * 7.0) * 0.14

        # Even harmonics (2, 4, 6) scaled by LFO
        sig = sig + SinOsc.ar(frequency=freq * 2.0) * 0.5 * even_mix
        sig = sig + SinOsc.ar(frequency=freq * 4.0) * 0.25 * even_mix
        sig = sig + SinOsc.ar(frequency=freq * 6.0) * 0.17 * even_mix

        sig = sig * 0.2  # normalize headroom
        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.5, sustain_time=5.0, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    synthdef = builder.build(name="additive")
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        synthdef.send(server)
        time.sleep(0.1)

        print("Playing additive synthesis with morphing harmonics (6.5s)...")
        server.synth("additive", frequency=150.0, amplitude=0.4)
        time.sleep(7.0)

    print("Done.")


if __name__ == "__main__":
    main()



"""
21_nodeproxy.py -- NodeProxy and Ndef live coding.

Demonstrates hot-swappable synth definitions via NodeProxy and Ndef.
The source is swapped multiple times while the monitor stays in place,
producing seamless audio transitions.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.proxy import Ndef, NodeProxy
from nanosynth.ugens import LFNoise1, LPF, Saw, SinOsc


def main() -> None:
    with Server(Options(verbosity=0, load_synthdefs=False)) as server:
        time.sleep(0.1)

        # -- NodeProxy: manual usage -------------------------------------------
        print("NodeProxy: sine -> saw -> clear")

        proxy = NodeProxy(server)
        proxy.source = lambda: SinOsc.ar(frequency=440) * 0.2
        proxy.play()
        print("  Playing sine at 440 Hz...")
        time.sleep(2.0)

        # Hot-swap to saw wave
        proxy.source = lambda: Saw.ar(frequency=330) * 0.15
        print("  Swapped to saw at 330 Hz...")
        time.sleep(2.0)

        # Hot-swap to filtered noise
        proxy.source = lambda: LPF.ar(
            source=LFNoise1.ar(frequency=800) * 0.2,
            frequency=600,
        )
        print("  Swapped to filtered noise...")
        time.sleep(2.0)

        proxy.clear()
        print("  Cleared.")
        time.sleep(0.3)

        # -- Ndef: named proxy registry ----------------------------------------
        print("\nNdef: named proxy registry")

        Ndef(server, "pad", lambda: SinOsc.ar(frequency=220) * 0.2)
        Ndef(server, "pad").play()
        print("  Ndef 'pad': sine 220 Hz")
        time.sleep(1.5)

        # Hot-swap via Ndef
        Ndef(server, "pad", lambda: Saw.ar(frequency=165) * 0.15)
        print("  Ndef 'pad': swapped to saw 165 Hz")
        time.sleep(1.5)

        # Second named proxy
        Ndef(server, "high", lambda: SinOsc.ar(frequency=880) * 0.1)
        Ndef(server, "high").play()
        print("  Ndef 'high': sine 880 Hz (layered)")
        time.sleep(2.0)

        # Clear all
        Ndef.clear_all(server)
        print("  Cleared all Ndefs.")
        time.sleep(0.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""
18_operators_buffers.py -- Extended operators and buffer management.

Demonstrates three features added in the unreleased version:

1. **Rich operator algebra** -- MIDI-to-frequency conversion (`midicps`),
   waveshaping (`tanh_`, `softclip`, `distort`), pitch math (`dbamp`),
   and signal clipping (`clip2`) applied directly on UGen signals.

2. **Buffer management** -- `managed_buffer` context manager for
   automatic allocation/deallocation, used here to back an FFT chain.

3. **Reply handling** -- `send_msg_sync` for synchronous buffer
   allocation (waits for `/done` before proceeding).

The synth plays a distorted saw chord with MIDI note inputs,
processed through FFT spectral filtering, then soft-clipped.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import (
    FFT,
    IFFT,
    LPF,
    Out,
    Pan2,
    PV_MagAbove,
    Saw,
    SinOsc,
)


def main() -> None:
    # -- SynthDef 1: operator showcase -----------------------------------------
    # Uses midicps (MIDI->Hz), tanh_ (soft saturation), clip2 (hard clip),
    # dbamp (dB->amplitude), and squared (spectral tilt).
    with SynthDefBuilder(note=60.0, detune=0.3, drive=4.0, amp_db=-12.0) as builder:
        # Convert MIDI note number to frequency using .midicps()
        freq = builder["note"].midicps()

        # Detuned saw pair for a thick sound
        sig = Saw.ar(frequency=freq)
        sig = sig + Saw.ar(frequency=freq * (1.0 + builder["detune"] * 0.01))

        # Drive into tanh_ waveshaper for warm saturation
        sig = (sig * builder["drive"]).tanh_()

        # Hard clip to +/- 0.8
        sig = sig.clip2(0.8)

        # Convert dB parameter to linear amplitude using .dbamp()
        amp = builder["amp_db"].dbamp()

        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.3, sustain_time=2.5, release_time=0.5
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * amp
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    operator_def = builder.build(name="op_demo")
    print(f"SynthDef '{operator_def.name}': {len(operator_def.compile())} bytes")
    print(operator_def.dump_ugens())

    # -- SynthDef 2: softclip + distort ----------------------------------------
    # Demonstrates .softclip() and .distort() waveshaping on a sine.
    with SynthDefBuilder(note=48.0, amplitude=0.25) as builder:
        freq = builder["note"].midicps()
        sig = SinOsc.ar(frequency=freq)
        # Overdrive then softclip: asymmetric warmth
        sig = (sig * 6.0).softclip()
        sig = sig + (SinOsc.ar(frequency=freq * 3.0) * 3.0).distort() * 0.3

        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.2, sustain_time=2.0, release_time=0.8
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=-0.3))

    softclip_def = builder.build(name="softclip_demo")
    print(f"\nSynthDef '{softclip_def.name}': {len(softclip_def.compile())} bytes")

    # -- SynthDef 3: spectral filtering with managed buffer --------------------
    # FFT processing using a server-allocated buffer (managed_buffer).
    with SynthDefBuilder(note=55.0, threshold=3.0, amplitude=0.3) as builder:
        freq = builder["note"].midicps()
        sig = Saw.ar(frequency=freq)
        sig = sig + Saw.ar(frequency=freq * 1.005)  # slight detune

        # FFT -> spectral gate -> IFFT
        chain = FFT.kr(source=sig)
        chain = PV_MagAbove.kr(pv_chain=chain, threshold=builder["threshold"])
        sig = IFFT.ar(pv_chain=chain)
        sig = LPF.ar(source=sig, frequency=5000.0)

        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.5, sustain_time=3.0, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=0.3))

    spectral_def = builder.build(name="spectral_op")
    print(f"SynthDef '{spectral_def.name}': {len(spectral_def.compile())} bytes")

    # -- Boot and play ---------------------------------------------------------
    with Server(Options(verbosity=0, load_synthdefs=False)) as server:
        operator_def.send(server)
        softclip_def.send(server)
        spectral_def.send(server)
        time.sleep(0.1)

        # --- Part 1: operator-heavy synths ---
        print("\n1. Operator showcase: MIDI notes -> midicps -> tanh_ -> clip2")
        # Play a chord: C3, E3, G3 as MIDI notes
        midi_notes = [48.0, 52.0, 55.0]
        for note in midi_notes:
            operator_def.play(server, note=note, detune=0.4, drive=3.0, amp_db=-14.0)
            time.sleep(0.15)
        print(f"   Playing MIDI chord {midi_notes} (3.5s)...")
        time.sleep(3.5)

        # --- Part 2: softclip/distort waveshaping ---
        print("2. Softclip + distort waveshaping on sine...")
        softclip_def.play(server, note=48.0, amplitude=0.2)
        time.sleep(3.5)

        # --- Part 3: buffer management + spectral processing ---
        print("3. Managed buffer + spectral gate...")

        # Allocate a buffer synchronously using send_msg_sync
        # (waits for /done reply before proceeding)
        buf_id = server.next_buffer_id()
        reply = server.send_msg_sync(
            "/b_alloc",
            buf_id,
            2048,
            1,
            reply_address="/done",
            timeout=2.0,
        )
        if reply:
            server._allocated_buffers.add(buf_id)
            print(f"   Buffer {buf_id} allocated synchronously (got /done reply)")
        else:
            print(f"   Buffer {buf_id} allocated (no reply -- fire-and-forget)")

        spectral_def.play(server, note=55.0, threshold=3.0, amplitude=0.25)
        time.sleep(4.5)

        # Clean up the manually allocated buffer
        server.free_buffer(buf_id)

        # Also demonstrate managed_buffer context manager
        with server.managed_buffer(num_frames=4096, num_channels=1) as mbuf:
            print(f"   Managed buffer {mbuf} allocated (auto-freed on exit)")
            spectral_def.play(server, note=62.0, threshold=2.0, amplitude=0.2)
            time.sleep(4.5)
        print("   Managed buffer freed.")

        time.sleep(0.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""Trigger and signal routing UGens."""

from ..enums import CalculationRate
from ..synthdef import (
    OutputProxy,
    UGen,
    UGenOperable,
    UGenRecursiveInput,
    UGenScalarInput,
    param,
    ugen,
)


@ugen(ar=True, kr=True, ir=True)
class Clip(UGen):
    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True, ir=True)
class Fold(UGen):
    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class Gate(UGen):
    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True, ir=True)
class InRange(UGen):
    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class Latch(UGen):
    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class LeastChange(UGen):
    a = param(0)
    b = param(0)


@ugen(ar=True, kr=True)
class MostChange(UGen):
    a = param(0)
    b = param(0)


@ugen(ar=True, kr=True)
class Peak(UGen):
    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class PeakFollower(UGen):
    source = param()
    decay = param(0.999)


@ugen(ar=True, kr=True)
class Phasor(UGen):
    trigger = param(0)
    rate = param(1.0)
    start = param(0.0)
    stop = param(1.0)
    reset_pos = param(0.0)


@ugen(ar=True, kr=True)
class RunningMax(UGen):
    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class RunningMin(UGen):
    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class Schmidt(UGen):
    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class SendTrig(UGen):
    trigger = param()
    id_ = param(0)
    value = param(0.0)


@ugen(ar=True, kr=True)
class Sweep(UGen):
    trigger = param(0)
    rate = param(1.0)


@ugen(ar=True, kr=True)
class TDelay(UGen):
    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True)
class ToggleFF(UGen):
    trigger = param(0)


@ugen(ar=True, kr=True)
class Trig1(UGen):
    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True)
class Trig(UGen):
    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True, ir=True)
class Wrap(UGen):
    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class ZeroCrossing(UGen):
    source = param()


@ugen(ar=True, kr=True)
class Poll(UGen):
    trigger = param()
    source = param()
    trigger_id = param(-1)
    label = param(unexpanded=True)

    def __init__(
        self,
        *,
        calculation_rate: CalculationRate = CalculationRate.SCALAR,
        label: UGenRecursiveInput | None = None,
        source: UGenScalarInput | None = None,
        trigger: UGenScalarInput | None = None,
        trigger_id: UGenScalarInput = -1,
        **kwargs: UGenRecursiveInput | None,
    ) -> None:
        if label is None:
            if isinstance(source, UGen):
                label_str = type(source).__name__
            elif isinstance(source, OutputProxy):
                label_str = type(source.ugen).__name__
            else:
                label_str = "UGen"
        else:
            label_str = str(label)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            label=[len(label_str), *(ord(c) for c in label_str)],
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )

    @classmethod
    def ar(
        cls,
        *,
        label: str | None = None,
        source: UGenRecursiveInput | None = None,
        trigger: UGenRecursiveInput | None = None,
        trigger_id: UGenRecursiveInput = -1,
    ) -> UGenOperable:
        return cls._new_expanded(
            calculation_rate=CalculationRate.AUDIO,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )

    @classmethod
    def kr(
        cls,
        *,
        label: str | None = None,
        source: UGenRecursiveInput | None = None,
        trigger: UGenRecursiveInput | None = None,
        trigger_id: UGenRecursiveInput = -1,
    ) -> UGenOperable:
        return cls._new_expanded(
            calculation_rate=CalculationRate.CONTROL,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class SendPeakRMS(UGen):
    reply_rate = param(20)
    peak_lag = param(3)
    reply_id = param(-1)
    source_size = param()
    source = param(unexpanded=True)
    character_count = param()
    character = param(unexpanded=True)

    @classmethod
    def ar(
        cls,
        *,
        command_name: str = "/reply",
        peak_lag: UGenScalarInput = 3,
        reply_id: UGenScalarInput = -1,
        reply_rate: UGenScalarInput = 20,
        source: UGenRecursiveInput,
    ) -> UGenOperable:
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.AUDIO,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
            source_size=len(source),  # type: ignore[arg-type]
            character_count=len(command),
            character=[ord(x) for x in command],
        )

    @classmethod
    def kr(
        cls,
        *,
        command_name: str = "/reply",
        peak_lag: UGenScalarInput = 3,
        reply_id: UGenScalarInput = -1,
        reply_rate: UGenScalarInput = 20,
        source: UGenRecursiveInput,
    ) -> UGenOperable:
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.CONTROL,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
            source_size=len(source),  # type: ignore[arg-type]
            character_count=len(command),
            character=[ord(x) for x in command],
        )


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class SendReply(UGen):
    trigger = param()
    reply_id = param(-1)
    character_count = param()
    character = param(unexpanded=True)
    source = param(unexpanded=True)

    @classmethod
    def ar(
        cls,
        *,
        command_name: str = "/reply",
        reply_id: UGenScalarInput = -1,
        source: UGenRecursiveInput,
        trigger: UGenRecursiveInput,
    ) -> UGenOperable:
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.AUDIO,
            trigger=trigger,
            reply_id=reply_id,
            character_count=len(command),
            character=[ord(x) for x in command],
            source=source,
        )

    @classmethod
    def kr(
        cls,
        *,
        command_name: str = "/reply",
        reply_id: UGenScalarInput = -1,
        source: UGenRecursiveInput,
        trigger: UGenRecursiveInput,
    ) -> UGenOperable:
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.CONTROL,
            trigger=trigger,
            reply_id=reply_id,
            character_count=len(command),
            character=[ord(x) for x in command],
            source=source,
        )



"""nanosynth -- minimal embedded SuperCollider synthesis engine wrapper."""

__version__ = "0.1.4"

from .enums import AddAction, CalculationRate, DoneAction
from .osc import OscBundle, OscMessage, find_free_port
from .scsynth import EmbeddedProcessProtocol, Options, find_ugen_plugins_path
from .compiler import compile_synthdefs
from .synthdef import (
    Default,
    Param,
    PseudoUGen,
    SynthDef,
    SynthDefBuilder,
    SynthDefGraph,
    UGen,
    UGenInput,
    UGenNode,
    control,
    param,
    synthdef,
    ugen,
)
from .envelopes import EnvGen, Envelope
from .patterns import (
    Clock,
    Pbind,
    Pchoose,
    Pconst,
    Pgeom,
    Pn,
    Prand,
    Pseq,
    Pseries,
    Pwhite,
    Pattern,
    Player,
    Rest,
)
from .proxy import Ndef, NodeProxy
from .score import Score
from .server import Bus, Group, Server, Synth
from .ugens import *  # noqa: F403

# Common UGens exported via star-import. The full set remains available
# via ``from nanosynth.ugens import *`` or qualified imports.
_COMMON_UGENS = [
    "BPF",
    "BrownNoise",
    "BufRd",
    "DelayL",
    "DelayN",
    "Dust",
    "FreeVerb",
    "HPF",
    "Impulse",
    "In",
    "LFNoise0",
    "LFNoise1",
    "LFNoise2",
    "LFPulse",
    "LFSaw",
    "LPF",
    "Line",
    "Mix",
    "Out",
    "Pan2",
    "PinkNoise",
    "PlayBuf",
    "Pulse",
    "RLPF",
    "Resonz",
    "Saw",
    "SinOsc",
    "WhiteNoise",
    "XLine",
]

__all__ = [
    "AddAction",
    "Bus",
    "CalculationRate",
    "Clock",
    "Default",
    "DoneAction",
    "EmbeddedProcessProtocol",
    "EnvGen",
    "Envelope",
    "Group",
    "Ndef",
    "NodeProxy",
    "OscBundle",
    "OscMessage",
    "Options",
    "Param",
    "Pattern",
    "Pbind",
    "Pchoose",
    "Pconst",
    "Pgeom",
    "Player",
    "Pn",
    "Prand",
    "Pseq",
    "Pseries",
    "PseudoUGen",
    "Pwhite",
    "Rest",
    "Score",
    "Server",
    "Synth",
    "SynthDef",
    "SynthDefBuilder",
    "SynthDefGraph",
    "UGen",
    "UGenInput",
    "UGenNode",
    "compile_synthdefs",
    "control",
    "find_free_port",
    "find_ugen_plugins_path",
    "param",
    "synthdef",
    "ugen",
] + _COMMON_UGENS



"""
03_fm.py -- FM synthesis + melody.

Two SynthDefs:
  1. FM synth with controllable carrier_freq, mod_ratio, mod_index, gate
     (uses Envelope.adsr for amplitude shaping)
  2. FM sweep with XLine-modulated index (self-freeing)

Plays a short melody with the gated FM synth, then fires a self-freeing sweep.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import Out, Pan2, SinOsc, XLine


def main() -> None:
    # -- SynthDef 1: gated FM synth -------------------------------------------
    with SynthDefBuilder(
        carrier_freq=440.0,
        mod_ratio=2.0,
        mod_index=3.0,
        amplitude=0.3,
        gate=1.0,
    ) as builder:
        mod_freq = builder["carrier_freq"] * builder["mod_ratio"]
        modulator = SinOsc.ar(frequency=mod_freq) * builder["mod_index"] * mod_freq
        carrier = SinOsc.ar(frequency=builder["carrier_freq"] + modulator)
        env = EnvGen.kr(
            envelope=Envelope.adsr(
                attack_time=0.01,
                decay_time=0.1,
                sustain=0.7,
                release_time=0.3,
            ),
            gate=builder["gate"],
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = carrier * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    fm_def = builder.build(name="fm_synth")
    print(f"SynthDef '{fm_def.name}' compiled: {len(fm_def.compile())} bytes")

    # -- SynthDef 2: self-freeing FM sweep ------------------------------------
    with SynthDefBuilder(carrier_freq=200.0, mod_ratio=3.0, amplitude=0.25) as builder:
        mod_index = XLine.kr(
            start=10.0,
            stop=0.1,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        mod_freq = builder["carrier_freq"] * builder["mod_ratio"]
        modulator = SinOsc.ar(frequency=mod_freq) * mod_index * mod_freq
        carrier = SinOsc.ar(frequency=builder["carrier_freq"] + modulator)
        sig = carrier * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    sweep_def = builder.build(name="fm_sweep")
    print(f"SynthDef '{sweep_def.name}' compiled: {len(sweep_def.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        fm_def.send(server)
        sweep_def.send(server)
        time.sleep(0.1)

        # Play a short melody with the gated FM synth
        melody = [
            (261.63, 0.3),  # C4
            (293.66, 0.3),  # D4
            (329.63, 0.3),  # E4
            (349.23, 0.3),  # F4
            (392.00, 0.5),  # G4
            (440.00, 0.5),  # A4
            (493.88, 0.5),  # B4
            (523.25, 0.8),  # C5
        ]

        print("Playing FM melody...")
        for freq, dur in melody:
            node = server.synth(
                "fm_synth",
                carrier_freq=freq,
                mod_ratio=2.0,
                mod_index=3.0,
                amplitude=0.25,
            )
            time.sleep(dur)
            # Release the gate to trigger the envelope release
            node.set(gate=0.0)
            time.sleep(0.05)

        time.sleep(0.5)

        # Fire the self-freeing FM sweep
        print("Playing FM sweep (4s)...")
        server.synth("fm_sweep", carrier_freq=150.0)
        time.sleep(4.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""Line and level UGens."""

from ..synthdef import (
    PseudoUGen,
    UGen,
    UGenOperable,
    UGenRecursiveInput,
    UGenVector,
    param,
    ugen,
)


@ugen(kr=True, is_pure=True)
class A2K(UGen):
    source = param()


@ugen(ar=True, ir=True, kr=True, is_pure=True)
class AmpComp(UGen):
    frequency = param(1000.0)
    root = param(0.0)
    exp = param(0.3333)


@ugen(ar=True, ir=True, kr=True, is_pure=True)
class AmpCompA(UGen):
    frequency = param(1000.0)
    root = param(0.0)
    min_amp = param(0.32)
    root_amp = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class DC(UGen):
    source = param()


@ugen(ar=True, is_pure=True)
class K2A(UGen):
    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class LinExp(UGen):
    source = param()
    input_minimum = param(0)
    input_maximum = param(1)
    output_minimum = param(1)
    output_maximum = param(2)


class LinLin(PseudoUGen):
    """Linear-to-linear mapping pseudo-UGen."""

    @staticmethod
    def ar(
        *,
        source: UGenRecursiveInput,
        input_minimum: UGenRecursiveInput = 0.0,
        input_maximum: UGenRecursiveInput = 1.0,
        output_minimum: UGenRecursiveInput = 1.0,
        output_maximum: UGenRecursiveInput = 2.0,
    ) -> UGenOperable:
        from .basic import MulAdd

        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)  # type: ignore[operator]
        offset = output_minimum - (scale * input_minimum)
        return MulAdd.new(source=source, multiplier=scale, addend=offset)  # type: ignore[attr-defined,no-any-return]

    @staticmethod
    def kr(
        *,
        source: UGenRecursiveInput,
        input_minimum: UGenRecursiveInput = 0.0,
        input_maximum: UGenRecursiveInput = 1.0,
        output_minimum: UGenRecursiveInput = 1.0,
        output_maximum: UGenRecursiveInput = 2.0,
    ) -> UGenOperable:
        from .basic import MulAdd

        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)  # type: ignore[operator]
        offset = output_minimum - (scale * input_minimum)
        return MulAdd.new(source=source, multiplier=scale, addend=offset)  # type: ignore[attr-defined,no-any-return]


@ugen(ar=True, kr=True, has_done_flag=True)
class Line(UGen):
    start = param(0.0)
    stop = param(1.0)
    duration = param(1.0)
    done_action = param(0)


class Silence(PseudoUGen):
    """Audio-rate silence pseudo-UGen."""

    @classmethod
    def ar(cls, channel_count: int = 1) -> UGenOperable:
        silence = DC.ar(source=0)  # type: ignore[attr-defined]
        if channel_count == 1:
            return silence  # type: ignore[no-any-return]
        return UGenVector(*([silence] * channel_count))


@ugen(ar=True, kr=True, has_done_flag=True)
class XLine(UGen):
    start = param(1.0)
    stop = param(2.0)
    duration = param(1.0)
    done_action = param(0)



"""Disk I/O UGens."""

from ..synthdef import UGen, param, ugen


@ugen(ar=True, is_multichannel=True, has_done_flag=True)
class DiskIn(UGen):
    buffer_id = param()
    loop = param(0)


@ugen(ar=True, channel_count=0, fixed_channel_count=True)
class DiskOut(UGen):
    buffer_id = param()
    source = param(unexpanded=True)


@ugen(ar=True, is_multichannel=True, has_done_flag=True)
class VDiskIn(UGen):
    buffer_id = param()
    rate = param(1)
    loop = param(0.0)
    send_id = param(0)



import os

from git import Repo
from utilities import assert_step, use_directory, run

class GitUtilities:
    @staticmethod
    def check_no_uncommitted_changes(repo: Repo) -> None:
        assert_step(not repo.bare, "There are uncommitted changes in git")

        # From https://stackoverflow.com/questions/31959425/how-to-get-staged-files-using-gitpython
        repo_name = GitUtilities.get_repo_name(repo)
        assert_step(len(repo.index.diff(None)) == 0, f"there are un-committed changes to {repo_name}")  # Modified
        assert_step(len(repo.index.diff("HEAD")) == 0, f"there are staged changes to {repo_name}")  # Staged

    @staticmethod
    def check_branch_name(repo: Repo, branch_name: str) -> None:
        repo_name = GitUtilities.get_repo_name(repo)
        assert_step((repo.active_branch.name == branch_name), f"the {repo_name} repo is not on branch {branch_name}")

    @staticmethod
    def get_repo_name(repo: Repo) -> str:
        repo_location = repo.working_dir
        repo_name = os.path.basename(repo_location)
        return repo_name

    @staticmethod
    def reset_and_clean_working_directory(project_dir: str) -> None:
        with use_directory(project_dir):
            # Delete untracked files:
            # - does not delete ignored files
            # - does not delete untracked files in new, untracked directories
            run(["git", "clean", "-f"])

            run(["git", "reset", "--hard"])

    @staticmethod
    def add_everything(directory: str) -> None:
        with use_directory(directory):
            run(["git", "add", "."])

    @staticmethod
    def commit_everything(directory: str, message: str) -> None:
        with use_directory(directory):
            run(["git", "commit", "-m", F"'{message}'"])

    @staticmethod
    def add_and_commit_everything(directory: str, message: str) -> None:
        GitUtilities.add_everything(directory)
        GitUtilities.commit_everything(directory, message)

    @staticmethod
    def push_active_branch_origin(directory: str) -> None:
        repo = Repo(directory)
        branch = repo.active_branch.name
        repo.remote('origin').push(branch)

    @staticmethod
    def pull_active_branch_origin(directory: str) -> None:
        repo = Repo(directory)
        branch = repo.active_branch.name
        repo.remote('origin').pull(branch)



"""
17_freqshift.py -- Frequency shifting and Hilbert transforms.

Bode frequency shifting moves all partials by a fixed Hz amount (unlike
pitch shifting which scales them). This breaks harmonic relationships,
producing bell-like or metallic timbres from harmonic inputs.

Sections:
  1. Slow sweep -- a saw wave with FreqShift sweeping 0 -> 100 Hz,
     gradually detuning the harmonic series into inharmonicity
  2. Barberpole phaser -- two frequency-shifted copies of a signal
     mixed with the original, shifted by +/- small amounts, creating
     an endlessly rising/falling phaser effect
  3. Ring mod territory -- larger shift amounts approaching ring
     modulation (FreqShift at high offset is single-sideband ring mod)

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import (
    FreqShift,
    LFNoise1,
    LPF,
    Line,
    Out,
    Pan2,
    Saw,
    SinOsc,
)


def main() -> None:
    # -- SynthDef 1: slow frequency shift sweep -------------------------------
    with SynthDefBuilder(frequency=220.0, amplitude=0.3) as builder:
        sig = Saw.ar(frequency=builder["frequency"])
        # Sweep shift amount from 0 (harmonic) to 100 Hz (inharmonic)
        shift = Line.kr(start=0.0, stop=100.0, duration=6.0)
        sig = FreqShift.ar(source=sig, frequency=shift)
        sig = LPF.ar(source=sig, frequency=6000.0)
        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.3, sustain_time=5.5, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    sweep_def = builder.build(name="freqshift_sweep")
    print(f"SynthDef '{sweep_def.name}' compiled: {len(sweep_def.compile())} bytes")

    # -- SynthDef 2: barberpole phaser ----------------------------------------
    with SynthDefBuilder(frequency=330.0, rate=0.5, amplitude=0.25) as builder:
        sig = Saw.ar(frequency=builder["frequency"])
        sig = sig + Saw.ar(frequency=builder["frequency"] * 0.501) * 0.5

        # Two opposite shifts create an endlessly rising/falling effect
        shift_amount = builder["rate"]
        up = FreqShift.ar(source=sig, frequency=shift_amount)
        down = FreqShift.ar(source=sig, frequency=shift_amount * -1.0)

        # Mix original + shifted for phaser-like interference
        mixed = sig * 0.5 + up * 0.3 + down * 0.3
        mixed = LPF.ar(source=mixed, frequency=5000.0)

        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.5, sustain_time=5.0, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        mixed = mixed * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=mixed))

    phaser_def = builder.build(name="barberpole")
    print(f"SynthDef '{phaser_def.name}' compiled: {len(phaser_def.compile())} bytes")

    # -- SynthDef 3: single-sideband ring modulation --------------------------
    with SynthDefBuilder(frequency=440.0, amplitude=0.2) as builder:
        sig = SinOsc.ar(frequency=builder["frequency"])
        # LFO-modulated shift for evolving metallic texture
        shift = LFNoise1.kr(frequency=0.3) * 200.0 + 300.0
        sig = FreqShift.ar(source=sig, frequency=shift)
        sig = LPF.ar(source=sig, frequency=8000.0)
        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.3, sustain_time=4.0, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    ringmod_def = builder.build(name="ssb_ringmod")
    print(f"SynthDef '{ringmod_def.name}' compiled: {len(ringmod_def.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0, load_synthdefs=False)) as server:
        sweep_def.send(server)
        phaser_def.send(server)
        ringmod_def.send(server)
        time.sleep(0.1)

        print("\n1. Frequency shift sweep: 0 -> 100 Hz (harmonic to inharmonic)...")
        sweep_def.play(server, frequency=220.0, amplitude=0.3)
        time.sleep(7.5)

        print("2. Barberpole phaser: endless rising/falling...")
        phaser_def.play(server, frequency=330.0, rate=0.5, amplitude=0.25)
        time.sleep(7.0)

        print("3. Single-sideband ring mod: wandering metallic texture...")
        ringmod_def.play(server, frequency=440.0, amplitude=0.25)
        time.sleep(6.0)

        time.sleep(0.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""
06_ringmod.py -- Ring modulation and amplitude modulation.

Two SynthDefs demonstrating different modulation techniques:
  1. Ring modulation: carrier * modulator (bipolar, produces sum and
     difference frequencies)
  2. Amplitude modulation: carrier * (modulator * 0.5 + 0.5) (unipolar,
     retains carrier fundamental with sidebands)

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import Options, Server
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFTri, Out, Pan2, SinOsc, XLine


def main() -> None:
    # -- SynthDef 1: ring modulation with sweeping mod frequency --------------
    with SynthDefBuilder(carrier_freq=440.0, amplitude=0.3) as builder:
        carrier = SinOsc.ar(frequency=builder["carrier_freq"])
        # Sweep modulator frequency from 2 Hz (tremolo) up to 300 Hz (sidebands)
        mod_freq = XLine.kr(
            start=2.0,
            stop=300.0,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        modulator = LFTri.ar(frequency=mod_freq)
        sig = carrier * modulator * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=-0.4))

    ring_def = builder.build(name="ring_mod")
    print(f"SynthDef '{ring_def.name}' compiled: {len(ring_def.compile())} bytes")

    # -- SynthDef 2: amplitude modulation with sweeping mod frequency ---------
    with SynthDefBuilder(carrier_freq=440.0, amplitude=0.3) as builder:
        carrier = SinOsc.ar(frequency=builder["carrier_freq"])
        mod_freq = XLine.kr(
            start=2.0,
            stop=300.0,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        # Unipolar modulator: 0..1 instead of -1..1
        modulator = SinOsc.ar(frequency=mod_freq) * 0.5 + 0.5
        sig = carrier * modulator * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=0.4))

    am_def = builder.build(name="am_mod")
    print(f"SynthDef '{am_def.name}' compiled: {len(am_def.compile())} bytes")

    # -- Boot and play --------------------------------------------------------
    with Server(Options(verbosity=0)) as server:
        ring_def.send(server)
        am_def.send(server)
        time.sleep(0.1)

        # Ring mod on left, AM on right -- hear the difference
        print("Ring modulation (left) -- sweeping mod freq 2->300 Hz (4s)...")
        server.synth("ring_mod", carrier_freq=440.0, amplitude=0.35)
        time.sleep(0.5)

        print("Amplitude modulation (right) -- same sweep...")
        server.synth("am_mod", carrier_freq=440.0, amplitude=0.35)
        time.sleep(4.5)

    print("Done.")


if __name__ == "__main__":
    main()



"""Bus input/output UGens."""

import itertools
from collections.abc import Sequence
from typing import Any

from ..enums import CalculationRate
from ..synthdef import UGen, UGenRecursiveInput, param, ugen


@ugen(ar=True, kr=True, is_multichannel=True)
class In(UGen):
    bus = param(0.0)


@ugen(ar=True, kr=True, is_multichannel=True)
class InFeedback(UGen):
    bus = param(0.0)


@ugen(ar=True, kr=True, is_multichannel=True)
class LocalIn(UGen):
    default = param(0.0, unexpanded=True)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        default = kwargs["default"]
        if not isinstance(default, Sequence):
            default = [default]  # type: ignore[list-item]
        defaults = [float(x) for x in default]  # type: ignore[arg-type]
        # Repeat defaults to fill channel_count
        result: list[float] = []
        cycle = itertools.cycle(defaults)
        for i, x in enumerate(cycle):
            if i >= self._channel_count:
                break
            result.append(x)
        kwargs["default"] = result
        return calculation_rate, kwargs


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class LocalOut(UGen):
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class OffsetOut(UGen):
    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class Out(UGen):
    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class ReplaceOut(UGen):
    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class XOut(UGen):
    bus = param(0)
    crossfade = param(0.0)
    source = param(unexpanded=True)



import os
import subprocess
import hashlib
import urllib.request, urllib.error

from typing import Callable, List, Any


def run(command: List[str]) -> None:
    print(command)
    subprocess.run(" ".join(command), shell=True, check=True)


def write_file(file_name: str, text: str) -> None:
    with open(file_name, 'w') as output:
        output.write(text)


def read_file(file_name: str) -> str:
    with open(file_name) as input:
        text = input.read()
    return text


def open_url(url: str) -> None:
    run(['open', url])


def read_url(url: str) -> str:
    file = urllib.request.urlopen(url)
    text = ''
    for line in file:
        decoded_line = line.decode("utf-8")
        text += decoded_line
    return text


def check_url_exists(url: str) -> bool:
    try:
        conn = urllib.request.urlopen(url)
    except urllib.error.HTTPError as e:
        # Return code error (e.g. 404, 501, ...)
        return False
    except urllib.error.URLError as e:
        # Not an HTTP-specific error (e.g. connection refused)
        return False
    else:
        # 200
        return True


def get_file_name(path: str) -> str:
    return os.path.split(path)[1]

pushstack = list()


def pushdir(dirname: str) -> None:
    global pushstack
    pushstack.append(os.getcwd())
    os.chdir(dirname)


def popdir() -> None:
    global pushstack
    os.chdir(pushstack.pop())


def use_directory(dir: str) -> Any: # work around for nested class
    class PushPopDirectory:
        def __init__(self, dir: str) -> None:
            self.dir = dir

        def __enter__(self) -> None:
            pushdir(dir)

        def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
            popdir()
    return PushPopDirectory(dir)


def replace_text_in_file(file_name: str, old_text: str, new_text: str) -> None:
    text = read_file(file_name)
    text = text.replace(old_text, new_text)
    write_file(file_name, text)


def calculate_sha256(file_name: str) -> str:
    with open(file_name, "rb") as f:
        bytes = f.read()  # read entire file as bytes
        readable_hash = hashlib.sha256(bytes).hexdigest()
        # print(readable_hash)
        return readable_hash


# condition should be an Expression
def assert_step(condition: bool, message:str = "") -> None:
    assert condition, message


def check_step(step: str) -> None:
    def do_nothing() -> None:
        pass
    check_step_with_revert(step, do_nothing)


def check_step_with_revert(step: str, revert_function: Callable) -> None:
    print("\nCHECK: ", step)
    response = input("  Press Y OR y to continue; Anything else to Quit: ")
    if response not in ['Y', 'y']:
        revert_function()
        exit(0)
    print('---------------------------------------------------------------')



#!/usr/bin/env python
from __future__ import print_function, absolute_import, unicode_literals

import difflib
import glob
import os
import re
import string
import subprocess
import sys
import threading
from argparse import ArgumentParser

# Whichcraft backported shutil.which implementation
# Taken from https://github.com/pydanny/whichcraft/blob/master/whichcraft.py (version 0.5.3)
#
# BEGIN BSD-LICENSED CODE
#
# Copyright (c) 2015-2016, Daniel Roy Greenfeld All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this list of conditions and
# the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
# and the following disclaimer in the documentation and/or other materials provided with the
# distribution.
#
# * Neither the name of whichcraft nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

try:  # Forced testing
    from shutil import which
except ImportError:  # Forced testing
    # Versions prior to Python 3.3 don't have shutil.which

    def which(cmd, mode=os.F_OK | os.X_OK, path=None):
        """Given a command, mode, and a PATH string, return the path which
        conforms to the given mode on the PATH, or None if there is no such
        file.
        `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
        of os.environ.get("PATH"), or can be overridden with a custom search
        path.
        Note: This function was backported from the Python 3 source code.
        """
        # Check that a given file can be accessed with the correct mode.
        # Additionally check that `file` is not a directory, as on Windows
        # directories pass the os.access check.

        def _access_check(fn, mode):
            return os.path.exists(fn) and os.access(fn, mode) and not os.path.isdir(fn)

        # If we're given a path with a directory part, look it up directly
        # rather than referring to PATH directories. This includes checking
        # relative to the current directory, e.g. ./script
        if os.path.dirname(cmd):
            if _access_check(cmd, mode):
                return cmd

            return None

        if path is None:
            path = os.environ.get("PATH", os.defpath)
        if not path:
            return None

        path = path.split(os.pathsep)

        if sys.platform == "win32":
            # The current directory takes precedence on Windows.
            if os.curdir not in path:
                path.insert(0, os.curdir)

            # PATHEXT is necessary to check on Windows.
            pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
            # See if the given file matches any of the expected path
            # extensions. This will allow us to short circuit when given
            # "python.exe". If it does match, only test that one, otherwise we
            # have to try others.
            if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
                files = [cmd]
            else:
                files = [cmd + ext for ext in pathext]
        else:
            # On other platforms you don't have things like PATHEXT to tell you
            # what file suffixes are executable, so just pass on cmd as-is.
            files = [cmd]

        seen = set()
        for dir in path:
            normdir = os.path.normcase(dir)
            if normdir not in seen:
                seen.add(normdir)
                for thefile in files:
                    name = os.path.join(dir, thefile)
                    if _access_check(name, mode):
                        return name

        return None

##############################################################################
# END BSD-LICENSED CODE
##############################################################################

##############################################################################
#
# Constants
#

CLANG_FORMAT_ACCEPTED_VERSION_REGEX = re.compile("14\\.\\d+\\.\\d+")
CLANG_FORMAT_ACCEPTED_VERSION_STRING = "14.y.z"

# all the extensions we format with clang-format in SC (no JS!)
CLANG_FORMAT_FILES_REGEX = re.compile('\\.(cpp|hpp|h|c|m|mm)$')

# autogen'd files, don't touch
AUTOGEN_FILES_REGEX = re.compile('(SCDoc\\.tab\\..pp|lex\\.scdoc\\.cpp|lang11d_tab\\..*)$')

# the destination filename for a git diff
DIFF_FILENAME_REGEX = re.compile('^\\+\\+\\+ b/(.*)$', re.MULTILINE)

##############################################################################

def callo(args):
    """Call a program, and capture its output
    """
    return subprocess.check_output(args).decode('utf-8')

def callo_as_bytes(args):
    """Call a program, and capture its output as bytes without decoding
    """
    return subprocess.check_output(args)

def callo_with_input(args, inputdata):
    """Call a program, pipe input into it, and capture its output
    """
    pipe = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    return pipe.communicate(inputdata.encode('utf-8'))[0].decode('utf-8')

def get_base_dir():
    """Get the base directory for repo.
    """
    try:
        return subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).rstrip().decode('utf-8')
    except:
        print("This script must be running in a git repo")
        sys.exit(2)

class Repo(object):
    """Class encapsulates all knowledge about a git repository, and its metadata
        to run clang-format.
    """
    def __init__(self, path):
        self.path = path

    def _callgito(self, args):
        """Call git for this repository, and return the captured output
        """
        # These two flags are the equivalent of -C in newer versions of Git
        # but we use these to support versions pre 1.8.5 but it depends on the command
        # and what the current directory is
        return callo(['git', '--git-dir', os.path.join(self.path, ".git"),
                            '--work-tree', self.path] + args)

    def _callgito_as_bytes(self, args):
        """Same as _callgito, but returns bytes instead of str, to prevent binary files from being decoded as utf8
        """
        return callo_as_bytes(['git', '--git-dir', os.path.join(self.path, ".git"),
            '--work-tree', self.path] + args)

    def _callgit(self, args, stdout=None):
        """Call git for this repository without capturing output
        This is designed to be used when git returns non-zero exit codes.
        """
        # These two flags are the equivalent of -C in newer versions of Git
        # but we use these to support versions pre 1.8.5 but it depends on the command
        # and what the current directory is
        return subprocess.call(['git', '--git-dir', os.path.join(self.path, ".git"),
                                '--work-tree', self.path] + args, stdout=stdout)

    def is_detached(self):
        # symbolic-ref returns 1 if the repo is in a detached HEAD state
        with open(os.devnull, 'w') as DEVNULL:
            return self._callgit(["symbolic-ref", "--quiet", "HEAD"], stdout=DEVNULL)

    def is_ancestor(self, parent, child):
        # merge base returns 0 if parent is an ancestor of child
        return not self._callgit(["merge-base", "--is-ancestor", parent, child])

    def is_commit(self, sha1):
        # cat-file -e returns 0 if it is a valid hash
        return not self._callgit(["cat-file", "-e", "%s^{commit}" % sha1])

    def is_working_tree_dirty(self):
        # diff returns 1 if the working tree has local changes
        return self._callgit(["diff", "--quiet"])

    def does_branch_exist(self, branch):
        # rev-parse returns 0 if the branch exists
        return not self._callgit(["rev-parse", "--verify", "--quiet", branch])

    def get_merge_base(self, commit):
        return self._callgito(["merge-base", "HEAD", commit]).rstrip()

    def get_branch_name(self):
        """Get the current branch name, short form
           This returns "main", not "refs/head/main"
           Will not work if the current branch is detached
        """
        branch = self.rev_parse(["--abbrev-ref", "HEAD"])
        if branch == "HEAD":
            raise ValueError("Branch is currently detached")

        return branch

    def add(self, command):       return self._callgito(["add"] + command)
    def checkout(self, command):  return self._callgito(["checkout"] + command)
    def commit(self, command):    return self._callgito(["commit"] + command)
    def diff(self, command):      return self._callgito(["diff"] + command)
    def log(self, command):       return self._callgito(["log"] + command)
    def rev_parse(self, command): return self._callgito(["rev-parse"] + command).rstrip()
    def rm(self, command):        return self._callgito(["rm"] + command)
    def lsfiles(self):            return self._callgito(["ls-files"])

    # uses bytes instead of decoding to protect binary files with non-utf8-decodable contents
    def show_as_bytes(self, command): return self._callgito_as_bytes(["show"] + command)

class ClangFormat(object):
    """Class encapsulates finding a suitable copy of clang-format,
    and linting/formating an individual file
    """
    def __init__(self, cf_cmd):
        self.cf_cmd = cf_cmd
        if which(cf_cmd) is None:
            raise ValueError("Could not find clang-format at %s" % cf_cmd)
        self._validate_version()

    def _validate_version(self):
        cf_version = callo([self.cf_cmd, "--version"])

        if CLANG_FORMAT_ACCEPTED_VERSION_REGEX.search(cf_version):
            return

        # TODO add instructions to check docs when docs are written
        raise ValueError("clang-format found, but incorrect version at " +
                self.cf_cmd + " with version: " + cf_version + "\nAccepted versions: " +
                CLANG_FORMAT_ACCEPTED_VERSION_STRING)
        sys.exit(5)

    def lint(self, file_name, print_diff):
        """Check the specified file has the correct format
        """
        with open(file_name, 'rb') as original_text:
            original_file = original_text.read().decode('utf-8')

        # Get formatted file as clang-format would format the file
        formatted_file = callo([self.cf_cmd, '-style=file', file_name])

        if original_file != formatted_file:
            if print_diff:
                original_lines = original_file.splitlines()
                formatted_lines = formatted_file.splitlines()
                result = difflib.unified_diff(original_lines, formatted_lines, file_name, file_name)
                for line in result:
                    print(line.rstrip())

            return False

        return True

    def format(self, file_name):
        """Update the format of the specified file
        """
        if self.lint(file_name, print_diff=False):
            return True

        # Update the file with clang-format
        formatted = not subprocess.call([self.cf_cmd, '-style=file', '-i', file_name])

        # Version 3.8 generates files like foo.cpp~RF83372177.TMP when it formats foo.cpp
        # on Windows, we must clean these up
        if sys.platform == "win32":
            glob_pattern = file_name + "*.TMP"
            for fglob in glob.glob(glob_pattern):
                os.unlink(fglob)

        return formatted

def get_list_from_lines(lines):
    """"Convert a string containing a series of lines into a list of strings
    """
    return [line.rstrip() for line in lines.splitlines()]

def validate_repo_state_for_rebase(commit_before_reformat, commit_after_reformat, target_branch):
    if sys.version_info[0] == 2:
        cwd = os.getcwdu()
    else:
        cwd = os.getcwd()

    if os.path.normpath(cwd) != os.path.normpath(get_base_dir()):
        raise ValueError("reformat-branch must be run from the repo root")

    repo = Repo(get_base_dir())

    if not repo.is_commit(commit_before_reformat):
        raise ValueError("Commit before reformat '%s' is not a valid commit in this repo" %
                commit_before_reformat)

    if not repo.is_commit(commit_after_reformat):
        raise ValueError("Commit after reformat '%s' is not a valid commit in this repo" %
                commit_after_reformat)

    if not repo.is_ancestor(commit_before_reformat, commit_after_reformat):
        raise ValueError(("Commit before reformat '%s' is not a valid ancestor of commit after" +
                " reformat '%s' in this repo") % (commit_before_reformat, commit_after_reformat))

    if repo.is_detached():
        raise ValueError("You must not run this script in a detached HEAD state")

    if repo.is_working_tree_dirty():
        raise ValueError("Your working tree has pending changes. You must have a clean working" +
            " tree before proceeding.\n\nRun `git status` to see your pending changes, and then" +
            " try `git stash save`, `git reset --hard`, `git submodule update` and/or committing" +
            " your changes.")

    merge_base = repo.get_merge_base(commit_before_reformat)

    if not merge_base == repo.rev_parse([commit_before_reformat]):
        raise ValueError(("Merge base is '%s'. Please rebase to '%s' and resolve all conflicts" +
            " before running this script.\n\nTo interactively rebase, use `git rebase -i %s`") %
            (merge_base, commit_before_reformat, commit_before_reformat))

    # We assume the target branch is main, it could be a different branch if needed for testing
    merge_base = repo.get_merge_base(target_branch)

    if not merge_base == repo.rev_parse([commit_before_reformat]):
        raise ValueError("This branch appears to already have advanced too far through the merge process")

    return repo

def get_branch_names(repo):
    # Everything looks good so lets start going through all the commits
    branch_name = repo.get_branch_name()
    new_branch = branch_name + "-reformatted"

    if repo.does_branch_exist(new_branch):
        raise ValueError("The branch '%s' already exists. Please delete the branch '%s', or rename the current branch." % (new_branch, new_branch))

    return (branch_name, new_branch)

def is_3rd_party_file(name):
    return name.find('external_libraries') != -1
def is_autogen_file(name):
    return AUTOGEN_FILES_REGEX.search(name)
def is_clang_formattable(name):
    return CLANG_FORMAT_FILES_REGEX.search(name)

def is_wanted_clang_formattable_file(f):
    """Is this something we want to use ClangFormat to format?
    """
    return is_clang_formattable(f) and not is_3rd_party_file(f) and not is_autogen_file(f)

def get_all_clang_formattable_files(repo):
    files = get_list_from_lines(repo.lsfiles())
    return [f for f in files if is_wanted_clang_formattable_file(f)]

def rebase_branch(clang_format, commit_before_reformat, commit_after_reformat, target_branch):
    """Reformat a branch made before a clang-format run
    """
    clang_format = ClangFormat(clang_format)
    repo = validate_repo_state_for_rebase(commit_before_reformat, commit_after_reformat, target_branch)
    old_branch, new_branch = get_branch_names(repo)
    commits = get_list_from_lines(repo.log(["--reverse", "--pretty=format:%H", "%s..HEAD" % commit_before_reformat]))
    previous_commit_base = commit_after_reformat

    # Go through all the commits the user made on the local branch and migrate to a new branch
    # that is based on post_reformat commits instead
    for idx, commit_hash in enumerate(commits):
        print("--- Formatting " + commit_hash + (" (%s of %s)" % (idx + 1, len(commits))))
        repo.checkout(["--quiet", "--detach", commit_hash])

        deleted_files = []

        # Format each of the files by checking out just a single commit from the user's branch
        commit_files = get_list_from_lines(repo.diff(["HEAD~", "--name-only"]))

        for commit_file in commit_files:

            # Format each file needed if it was not deleted
            if not os.path.exists(commit_file):
                print("\tSkipping file '%s' since it has been deleted in commit '%s'" % (
                        commit_file, commit_hash))
                deleted_files.append(commit_file)
                continue

            if is_3rd_party_file(commit_file):
                print("\tSkipping external libraries file '%s'" % commit_file)
            elif is_autogen_file(commit_file):
                print("\tSkipping autogenerated file '%s'" % commit_file)
            elif is_clang_formattable(commit_file):
                clang_format.format(commit_file)
            else:
                print("\tSkipping file '%s' (no formatting to apply)" % commit_file)

        # Check if anything needed reformatting, and if so amend the commit
        if not repo.is_working_tree_dirty():
            print ("Commit %s needed no reformatting" % commit_hash)
        else:
            repo.commit(["--all", "--amend", "--no-edit"])

        # Rebase our new commit on top the post-reformat commit
        previous_commit = repo.rev_parse(["HEAD"])

        # Checkout the new branch with the reformatted commits
        # Note: we will not name as a branch until we are done with all commits on the local branch
        repo.checkout(["--quiet", "--detach", previous_commit_base])

        # Copy each file from the reformatted commit on top of the post reformat
        diff_files = get_list_from_lines(repo.diff(["%s~..%s" % (previous_commit, previous_commit),
            "--name-only"]))

        for diff_file in diff_files:
            # If the file was deleted in the commit we are reformatting, we need to delete it again
            if diff_file in deleted_files:
                repo.rm([diff_file])
                continue

            # The file has been added or modified, continue as normal
            # Get file as bytes to prevent utf8-decoding of binary files
            file_contents = repo.show_as_bytes(["%s:%s" % (previous_commit, diff_file)])

            root_dir = os.path.dirname(diff_file)
            if root_dir and not os.path.exists(root_dir):
                os.makedirs(root_dir)

            with open(diff_file, "bw+") as new_file:
                new_file.write(file_contents)

            repo.add([diff_file])

        # Create a new commit onto clang-formatted branch
        repo.commit(["--reuse-message=%s" % previous_commit])

        previous_commit_base = repo.rev_parse(["HEAD"])

    # Create a new branch to mark the hashes we have been using
    repo.checkout(["-b", new_branch])

    print("reformat-branch is done running.\n")
    print("A copy of your branch has been made named '%s', and formatted with clang-format.\n" % new_branch)
    print("The original branch has been left unchanged.")
    print("If you have not just done so, the next step is to rebase the new branch on '%s'.\n" % target_branch)
    print("To undo this, run `git checkout %s && git branch -D %s`" % (old_branch, new_branch))

def is_wanted_diff(diff_text):
    # Extract file name
    match = DIFF_FILENAME_REGEX.search(diff_text)
    if not match:
        if '+++ /dev/null' in diff_text:
            # The file was deleted, so ignore it:
            return False;
        raise ValueError("Could not extract filename from diff")
    return is_wanted_clang_formattable_file(match.group(1))

def filter_unwanted_files_from_diff(diff_text):
    # git diff was called with -U0 so all actual diffed lines can't start with '^diff'
    # Couldn't find a way to split on lookaheads, so went with this instead.
    # [1:] to discard initial empty string
    diffs = ['diff' + match for match in re.split('^diff', diff_text, flags=re.MULTILINE)][1:]
    filter_diffs = [diff for diff in diffs if is_wanted_diff(diff)]
    return ''.join(filter_diffs)

def prepare_diff_for_lint_format(clang_format, commit):
    ClangFormat(clang_format) # validation

    repo = Repo(get_base_dir())
    if not repo.is_commit(commit):
        raise ValueError("Commit before reformat '%s' is not a valid commit in this repo" % commit)

    os.chdir(repo.path)

    diff_text = repo.diff([commit, '-U0', '--no-color'])
    return filter_unwanted_files_from_diff(diff_text)

def do_lint(clang_format, clang_format_diff, commit):
    diff_text = prepare_diff_for_lint_format(clang_format, commit)
    lint_out = callo_with_input(['python', clang_format_diff, '-p1', '-binary', clang_format], diff_text)
    print(lint_out, end='')
    if lint_out != '\n' and lint_out != '':
        sys.exit(1)

def do_format(clang_format, clang_format_diff, commit):
    diff_text = prepare_diff_for_lint_format(clang_format, commit)
    callo_with_input(['python', clang_format_diff, '-i', '-p1', '-binary', clang_format], diff_text)

def do_lintall(clang_format):
    repo = Repo(get_base_dir())
    os.chdir(repo.path)
    clang_format = ClangFormat(clang_format)
    no_changes_needed = True
    for f in get_all_clang_formattable_files(repo):
        no_changes_needed = clang_format.lint(f, True) and no_changes_needed
    if not no_changes_needed:
        sys.exit(1)

def do_formatall(clang_format):
    repo = Repo(get_base_dir())
    os.chdir(repo.path)
    clang_format = ClangFormat(clang_format)
    for f in get_all_clang_formattable_files(repo):
        clang_format.format(f)

def resolve_program_name(cmd_line_option, env_var_name, default_program_name):
    if cmd_line_option != '':
        return cmd_line_option
    elif env_var_name in os.environ and os.environ[env_var_name] != '':
        return os.environ[env_var_name]
    else:
        return default_program_name

def main():
    parser = ArgumentParser(
            usage='''
    format.py lint [commit]
    format.py format [commit]
    format.py lintall
    format.py formatall
    format.py rebase -b base-branch
    format.py rebase commit1 commit2 target

PLEASE READ.

This script provides commands for linting and formatting your working directory. It provides five
commands:
1. `lint` lints the diff between the working directory and a given commit
2. `format` will apply formatting rules to the diff between working directory and given commit
3. `lintall` lints all available files for various formatting rules and indicates any problems.
4. `formatall` formats all available files.
5. `rebase` reformats a branch past the great reformatting wall. It can be run two ways; the second
   is simpler and usually works.
    a. `format.py rebase commit-right-before-reformat commit-after-reformat original-branch`
    b. `format.py 3.10 # or develop`

Rebase requires:
- you have a clean working directory
- you have rebased your branch on commit-right-before-reformat (implicitly for the second usage)
- you have the branch you want to rebase currently checked out

If there is an issue, this script will most likely detect it and provide you with commands to
proceed.

'commit' arguments can be a branch name, tag, or commit hash.

This script will exit with 0 on success, 1 to indicate lint failure, and >1 if some other error
occurs.
''')
    parser.add_argument("-c", "--clang-format", dest="clang_format", default='',
            help='Command to use for clang-format; will also be passed to clang-format-diff.py.'
            + ' Defaults to environment variable SC_CLANG_FORMAT if it is set and non-empty,'
            + ' otherwise `clang-format`')
    parser.add_argument("-b", "--base", dest="base_branch", help='Tries to rebase on the tip of this'
            + ' branch given a base branch name (experimental). This should be the main branch the'
            + ' current branch is based on (3.10 or develop)')
    parser.add_argument("-d", "--clang-format-diff", dest="clang_format_diff", default='',
            help='Command to use for clang-format-diff.py script'
            + ' Defaults to environment variable SC_CLANG_FORMAT_DIFF if it is set and non-empty,'
            + ' otherwise `clang-format-diff.py`')
    parser.add_argument("command", help="command; one of lint, format, lintall, formatall, rebase")
    parser.add_argument("commit1", help="for lint and format: commit to compare against (default: HEAD);" +
            " for rebase: commit immediately prior to reformat", nargs='?', default='')
    parser.add_argument("commit2", help="commit after reformat", nargs='?', default='')
    parser.add_argument("target", help="target branch name (likely 3.10 or develop)", nargs='?', default='')

    options = parser.parse_args()

    options.clang_format = resolve_program_name(options.clang_format, 'SC_CLANG_FORMAT', 'clang-format')
    options.clang_format_diff = resolve_program_name(options.clang_format_diff, 'SC_CLANG_FORMAT_DIFF', 'clang-format-diff.py')

    # in CI we want to replace the 'lint' and 'format' command with a full scan
    # to obtain a full stateless scan of the repository.
    # We do this by passing an env variable, see `.pre-commit-config.yaml`
    # and github actions workflow
    if os.environ.get("FULL_CHECK") is not None:
        print("Found FULL_CHECK environment variable. Perform a full scan")
        if options.command == 'lint':
            options.command = 'lintall'
        elif options.command == 'format':
            options.command = 'formatall'
        else:
            print("FULL_CHECK is only compatible with 'format' and 'lint' commands")
            sys.exit(1)

    try:
        if options.command == 'lint' or options.command == 'format':
            commit = 'HEAD' if options.commit1 == '' else options.commit1

            # For portability, we use the full path of the clang-format-diff.py script. subprocess
            # module on Windows won't be able to find a Python-executable python script in PATH, and
            # if we invoke it with `python <script> <args>` then the python interpreter needs the
            # full path of the script. Of course, the downside is that we use whatever `python`
            # resolves to in the host system's shell.
            clang_format_diff_path = which(options.clang_format_diff)
            if clang_format_diff_path is None:
                if options.clang_format_diff == 'clang-format-diff.py':
                    raise ValueError(
                            "Could not find clang-format-diff.py. "
                            "Please ensure that clang %s is installed and that "
                            "clang-format-diff.py is in your PATH."
                            % CLANG_FORMAT_ACCEPTED_VERSION_STRING)
                else:
                    raise ValueError("Could not find clang-format-diff.py at %s." % options.clang_format_diff)
            if options.command == 'lint':
                do_lint(options.clang_format, clang_format_diff_path, commit)
            else:
                do_format(options.clang_format, clang_format_diff_path, commit)
        elif options.command == 'lintall':
            do_lintall(options.clang_format)
        elif options.command == 'formatall':
            do_formatall(options.clang_format)
        elif options.command == 'rebase':
            if not options.commit1 or not options.commit2 or not options.target:
                if not options.base_branch:
                    parser.print_help()
                    sys.exit(2)

                if options.base_branch == '3.10':
                    options.commit1 = 'tag-clang-format-3.10^'
                    options.commit2 = options.target = 'tag-clang-format-3.10'
                elif options.base_branch == 'develop':
                    options.commit1 = 'tag-clang-format-develop^'
                    options.commit2 = options.target = 'tag-clang-format-develop'
                else:
                    print("Don't know how to use this base branch: %s. Try using the three-argument " +
                            "version of rebase command")
                    sys.exit(3)

            rebase_branch(options.clang_format, options.commit1, options.commit2, options.target)
        else:
            parser.print_help()
            sys.exit(4)
    except ValueError as ve:
        # print entire traceback to aid in diagnosing issues
        import traceback
        traceback.print_tb(sys.exc_info()[2])
        print("\n*** ERROR:\n" + str(ve) + "\n")
        sys.exit(6)

if __name__ == "__main__":
    main()



"""BEQ (BiQuad EQ) filter UGens."""

from ..synthdef import UGen, param, ugen


@ugen(ar=True, is_pure=True)
class BAllPass(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BBandPass(UGen):
    source = param()
    frequency = param(1200.0)
    bandwidth = param(1.0)


@ugen(ar=True, is_pure=True)
class BBandStop(UGen):
    source = param()
    frequency = param(1200.0)
    bandwidth = param(1.0)


@ugen(ar=True, is_pure=True)
class BHiCut(UGen):
    source = param()
    frequency = param(1200.0)
    order = param(2.0)
    max_order = param(5.0)


@ugen(ar=True, is_pure=True)
class BHiPass(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BHiShelf(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_s = param(1.0)
    gain = param(0.0)


@ugen(ar=True, is_pure=True)
class BLowCut(UGen):
    source = param()
    frequency = param(1200.0)
    order = param(2.0)
    max_order = param(5.0)


@ugen(ar=True, is_pure=True)
class BLowPass(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, is_pure=True)
class BLowShelf(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_s = param(1.0)
    gain = param(0.0)


@ugen(ar=True, is_pure=True)
class BPeakEQ(UGen):
    source = param()
    frequency = param(1200.0)
    reciprocal_of_q = param(1.0)
    gain = param(0.0)



"""Buffer I/O UGens."""

from typing import Any

from ..enums import CalculationRate
from ..synthdef import (
    Default,
    UGen,
    UGenRecursiveInput,
    param,
    ugen,
)


@ugen(ar=True, kr=True, is_multichannel=True)
class BufRd(UGen):
    buffer_id = param()
    phase = param(0.0)
    loop = param(1)
    interpolation = param(2)


@ugen(ar=True, kr=True, has_done_flag=True)
class BufWr(UGen):
    buffer_id = param()
    phase = param(0.0)
    loop = param(1.0)
    source = param(unexpanded=True)


@ugen(ir=True, is_width_first=True)
class ClearBuf(UGen):
    buffer_id = param()


@ugen(ir=True, is_width_first=True)
class LocalBuf(UGen):
    channel_count = param(1)
    frame_count = param(1)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        return CalculationRate.SCALAR, kwargs


@ugen(ir=True)
class MaxLocalBufs(UGen):
    maximum = param(0)


@ugen(ar=True, kr=True, is_multichannel=True)
class PlayBuf(UGen):
    buffer_id = param()
    rate = param(1)
    trigger = param(1)
    start_position = param(0)
    loop = param(0)
    done_action = param(0)


@ugen(ar=True, kr=True, has_done_flag=True)
class RecordBuf(UGen):
    buffer_id = param()
    offset = param(0.0)
    record_level = param(1.0)
    preexisting_level = param(0.0)
    run = param(1.0)
    loop = param(1.0)
    trigger = param(1.0)
    done_action = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True)
class ScopeOut(UGen):
    buffer_id = param()
    source = param(unexpanded=True)


@ugen(ar=True, kr=True)
class ScopeOut2(UGen):
    scope_id = param()
    max_frames = param(4096)
    scope_frames = param(Default())
    source = param(unexpanded=True)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        if isinstance(kwargs["scope_frames"], Default):
            kwargs["scope_frames"] = kwargs["max_frames"]
        return calculation_rate, kwargs
